# -*- coding: utf-8 -*-
# !/usr/bin/env python
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import csv

from oslo_config import cfg

from cloudkitty import tenant_fetcher

fake_fetcher_opts = [
    cfg.StrOpt('file',
               default='/var/lib/cloudkitty/tenants.csv',
               help='Fetcher input file.')]

cfg.CONF.register_opts(fake_fetcher_opts, 'fake_fetcher')
CONF = cfg.CONF


class FakeFetcher(tenant_fetcher.BaseFetcher):
    """Fake tenants fetcher."""

    def __init__(self):
        filename = cfg.CONF.fake_fetcher.file
        csvfile = open(filename, 'rb')
        reader = csv.DictReader(csvfile)
        self._csv = reader

    def get_tenants(self):
        return [row['id'] for row in self._csv]



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#
import decimal

from keystoneauth1 import loading as ks_loading
from keystoneclient.v3 import client as ks_client
from monascaclient import client as mclient
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from cloudkitty import collector
from cloudkitty import transformer
from cloudkitty import utils as ck_utils


LOG = logging.getLogger(__name__)

MONASCA_API_VERSION = '2_0'
COLLECTOR_MONASCA_OPTS = 'collector_monasca'
collector_monasca_opts = ks_loading.get_auth_common_conf_options()

cfg.CONF.register_opts(collector_monasca_opts, COLLECTOR_MONASCA_OPTS)
ks_loading.register_session_conf_options(
    cfg.CONF,
    COLLECTOR_MONASCA_OPTS)
ks_loading.register_auth_conf_options(
    cfg.CONF,
    COLLECTOR_MONASCA_OPTS)
CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)


class EndpointNotFound(Exception):
    """Exception raised if the Monasca endpoint is not found"""
    pass


class MonascaCollector(collector.BaseCollector):
    collector_name = 'monasca'
    dependencies = ['CloudKittyFormatTransformer']
    retrieve_mappings = {
        'compute': 'cpu',
        'image': 'image.size',
        'volume': 'volume.size',
        'network.floating': 'ip.floating',
        'network.bw.in': 'network.incoming.bytes',
        'network.bw.out': 'network.outgoing.bytes',
    }
    metrics_mappings = {
        'compute': [
            {'cpu': 'max'},
            {'vpcus': 'max'},
            {'memory': 'max'}],
        'image': [
            {'image.size': 'max'},
            {'image.download': 'max'},
            {'image.serve': 'max'}],
        'volume': [
            {'volume.size': 'max'}],
        'network.bw.in': [
            {'network.incoming.bytes': 'max'}],
        'network.bw.out': [
            {'network.outgoing.bytes': 'max'}],
        'network.floating': [
            {'ip.floating': 'max'}],
    }
    # (qty, unit). qty must be either a metric name, an integer
    # or a decimal.Decimal object
    units_mappings = {
        'compute': (1, 'instance'),
        'image': ('image.size', 'MiB'),
        'volume': ('volume.size', 'GiB'),
        'network.bw.out': ('network.outgoing.bytes', 'MB'),
        'network.bw.in': ('network.incoming.bytes', 'MB'),
        'network.floating': (1, 'ip'),
    }
    default_unit = (1, 'unknown')

    def __init__(self, transformers, **kwargs):
        super(MonascaCollector, self).__init__(transformers, **kwargs)

        self.t_cloudkitty = self.transformers['CloudKittyFormatTransformer']

        self.auth = ks_loading.load_auth_from_conf_options(
            CONF,
            COLLECTOR_MONASCA_OPTS)
        self.session = ks_loading.load_session_from_conf_options(
            CONF,
            COLLECTOR_MONASCA_OPTS,
            auth=self.auth)
        self.ks_client = ks_client.Client(session=self.session)
        self.mon_endpoint = self._get_monasca_endpoint()
        if not self.mon_endpoint:
            raise EndpointNotFound()
        # NOTE (lukapeschke) session authentication should be possible starting
        # with OpenStack Q release.
        self._conn = mclient.Client(
            api_version=MONASCA_API_VERSION,
            session=self.session,
            endpoint=self.mon_endpoint)

    # NOTE(lukapeschke) This function should be removed as soon as the endpoint
    # it no longer required by monascaclient
    def _get_monasca_endpoint(self, service_name='monasca',
                              endpoint_interface_type='public'):
        service_list = self.ks_client.services.list(name=service_name)
        if not service_list:
            return None
        mon_service = service_list[0]
        endpoints = self.ks_client.endpoints.list(mon_service.id)
        for endpoint in endpoints:
            if endpoint.interface == endpoint_interface_type:
                return endpoint.url
        return None

    def _get_metadata(self, resource_type, transformers):
        info = {}
        try:
            met = list(METRICS_CONF['metrics_units'][resource_type].values())
            info['unit'] = met[0]['unit']
        # NOTE(mc): deprecated second try kept for backward compatibility.
        except KeyError:
            LOG.warning('Error when trying to use yaml metrology conf.')
            LOG.warning('Fallback on the deprecated oslo config method.')
            try:
                info['unit'] = self.units_mappings[resource_type][1]
            except (KeyError, IndexError):
                info['unit'] = self.default_unit[1]

        start = ck_utils.dt2ts(ck_utils.get_month_start())
        end = ck_utils.dt2ts(ck_utils.get_month_end())

        try:
            resource = self.active_resources(resource_type, start,
                                             end, None)[0]
        except IndexError:
            resource = {}
        info['metadata'] = resource.get('dimensions', {}).keys()

        try:
            service_metrics = METRICS_CONF['services_metrics'][resource_type]
            for service_metric in service_metrics:
                metric, statistics = list(service_metric.items())[0]
                info['metadata'].append(metric)
        # NOTE(mc): deprecated second try kept for backward compatibility.
        except KeyError:
            LOG.warning('Error when trying to use yaml metrology conf.')
            LOG.warning('Fallback on the deprecated oslo config method.')
            try:
                for metric, statistics in self.metrics_mappings[resource_type]:
                    info['metadata'].append(metric)
            except (KeyError, IndexError):
                pass
        return info

    # NOTE(lukapeschke) if anyone sees a better way to do this,
    # please make a patch
    @classmethod
    def get_metadata(cls, resource_type, transformers):
        args = {
            'transformers': transformer.get_transformers(),
            'period': CONF.collect.period}
        tmp = cls(**args)
        return tmp._get_metadata(resource_type, transformers)

    def _get_resource_qty(self, meter, start, end, resource_id, statistics):
        # NOTE(lukapeschke) the period trick is used to aggregate
        # the measurements
        period = end - start
        statistics = self._conn.metrics.list_statistics(
            name=meter,
            start_time=ck_utils.ts2dt(start),
            end_time=ck_utils.ts2dt(end),
            dimensions={'resource_id': resource_id},
            statistics=statistics,
            period=period,
            merge_metrics=True,
        )
        try:
            # If several statistics are returned (should not happen),
            # use the latest
            qty = decimal.Decimal(statistics[-1]['statistics'][-1][1])
        except (KeyError, IndexError):
            qty = decimal.Decimal(0)
        return qty

    def _is_resource_active(self, meter, resource_id, start, end):
        measurements = self._conn.metrics.list_measurements(
            name=meter,
            start_time=ck_utils.ts2dt(start),
            end_time=ck_utils.ts2dt(end),
            group_by='resource_id',
            merge_metrics=True,
            dimensions={'resource_id': resource_id},
        )
        return len(measurements) > 0

    def active_resources(self, resource_type, start,
                         end, project_id, **kwargs):
        try:
            meter = METRICS_CONF['services_objects'].get(resource_type)
        # NOTE(mc): deprecated except part kept for backward compatibility.
        except KeyError:
            LOG.warning('Error when trying to use yaml metrology conf.')
            LOG.warning('Fallback on the deprecated oslo config method.')
            meter = self.retrieve_mappings.get(resource_type)

        if not meter:
            return {}
        dimensions = {}
        if project_id:
            dimensions['project_id'] = project_id
        dimensions.update(kwargs)
        resources = self._conn.metrics.list(name=meter, dimensions=dimensions)
        output = []
        for resource in resources:
            try:
                resource_id = resource['dimensions']['resource_id']
                if (resource_id not in
                    [item['dimensions']['resource_id'] for item in output]
                        and self._is_resource_active(meter, resource_id,
                                                     start, end)):
                    output.append(resource)
            except KeyError:
                continue
        return output

    def _expand_metrics(self, resource, resource_id,
                        mappings, start, end, resource_type):
        for mapping in mappings:
            name, statistics = list(mapping.items())[0]
            qty = self._get_resource_qty(
                name,
                start,
                end,
                resource_id,
                statistics,
            )

            try:
                conv_data = METRICS_CONF['metrics_units'][resource_type]
                conv_data = conv_data.get(name)
                if conv_data:
                    resource[name] = ck_utils.convert_unit(
                        qty,
                        conv_data.get('factor', 1),
                        conv_data.get('offset', 0),
                    )
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning(
                    'Error when trying to use yaml metrology conf.\n'
                    'Fallback on the deprecated hardcoded dict method.')

                names = ['network.outgoing.bytes', 'network.incoming.bytes']
                if name in names:
                    qty = qty / units.M
                elif 'image.' in name:
                    qty = qty / units.Mi
                resource[name] = qty

    def resource_info(self, resource_type, start, end,
                      project_id, q_filter=None):

        try:
            tmp = METRICS_CONF['metrics_units'][resource_type]
            qty = list(tmp.keys())[0]
            unit = list(tmp.values())[0]['unit']
        # NOTE(mc): deprecated except part kept for backward compatibility.
        except KeyError:
            LOG.warning('Error when trying to use yaml metrology conf.')
            LOG.warning('Fallback on the deprecated oslo config method.')
            qty, unit = self.units_mappings.get(
                resource_type,
                self.default_unit
            )

        active_resources = self.active_resources(
            resource_type, start, end, project_id
        )
        resource_data = []
        for resource in active_resources:
            resource_id = resource['dimensions']['resource_id']
            data = resource['dimensions']
            try:
                mappings = METRICS_CONF['services_metrics'][resource_type]
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                mappings = self.metrics_mappings[resource_type]

            self._expand_metrics(
                data,
                resource_id,
                mappings,
                start,
                end,
                resource_type,
            )
            resource_qty = qty
            if not (isinstance(qty, int) or isinstance(qty, decimal.Decimal)):
                try:
                    resource_qty \
                        = METRICS_CONF['services_objects'][resource_type]
                # NOTE(mc): deprecated except part kept for backward compat.
                except KeyError:
                    LOG.warning('Error when trying to use yaml metrology conf')
                    msg = 'Fallback on the deprecated oslo config method'
                    LOG.warning(msg)
                    resource_qty = data[self.retrieve_mappings[resource_type]]
                resource_qty = data[resource_qty]

            resource = self.t_cloudkitty.format_item(data, unit, resource_qty)
            resource['desc']['resource_id'] = resource_id
            resource['resource_id'] = resource_id
            resource_data.append(resource)
        return resource_data

    def retrieve(self, resource_type, start, end, project_id, q_filter=None):
        resources = self.resource_info(resource_type, start, end,
                                       project_id=project_id,
                                       q_filter=q_filter)
        if not resources:
            raise collector.NoDataCollected(self.collector_name, resource_type)
        return self.t_cloudkitty.format_service(resource_type, resources)



# -*- coding: utf-8 -*-
# Copyright 2017 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Luka Peschke
#
import abc

import six


@six.add_metaclass(abc.ABCMeta)
class BaseHybridBackend(object):
    """Base Backend class for the Hybrid Storage.

    This is the interface that all backends for the hybrid storage
    should implement.
    """

    @abc.abstractmethod
    def commit(self, tenant_id, state):
        """Push data to the storage backend.

        :param tenant_id: id of the tenant which information must be commited.
        """
        pass

    @abc.abstractmethod
    def init(self):
        """Initialize hybrid storage backend.

        Can be used to create DB scheme on first start
        """
        pass

    @abc.abstractmethod
    def get_total(self, begin=None, end=None, tenant_id=None,
                  service=None, groupby=None):
        """Return the current total.

        :param begin: When to start filtering.
        :type begin: datetime.datetime
        :param end: When to stop filtering.
        :type end: datetime.datetime
        :param tenant_id: Filter on the tenant_id.
        :type res_type: str
        :param service: Filter on the resource type.
        :type service: str
        :param groupby: Fields to group by, separated by commas if multiple.
        :type groupby: str
        """
        pass

    @abc.abstractmethod
    def append_time_frame(self, res_type, frame, tenant_id):
        """Append a time frame to commit to the backend.

        :param res_type: The resource type of the dataframe.
        :param frame: The timeframe to append.
        :param tenant_id: Tenant the frame is belonging to.
        """
        pass

    @abc.abstractmethod
    def get_tenants(self, begin, end):
        """Return the list of rated tenants.

        :param begin: When to start filtering.
        :type begin: datetime.datetime
        :param end: When to stop filtering.
        :type end: datetime.datetime
        """

    @abc.abstractmethod
    def get_time_frame(self, begin, end, **filters):
        """Request a time frame from the storage backend.

        :param begin: When to start filtering.
        :type begin: datetime.datetime
        :param end: When to stop filtering.
        :type end: datetime.datetime
        :param res_type: (Optional) Filter on the resource type.
        :type res_type: str
        :param tenant_id: (Optional) Filter on the tenant_id.
        :type res_type: str
        """



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import pecan
import six
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1 import types as ck_types
from cloudkitty import rating
from cloudkitty.rating.hash.controllers import field as field_api
from cloudkitty.rating.hash.controllers import mapping as mapping_api
from cloudkitty.rating.hash.datamodels import service as service_models
from cloudkitty.rating.hash.db import api as db_api


class HashMapServicesController(rating.RatingRestControllerBase):
    """Controller responsible of services management.

    """

    fields = field_api.HashMapFieldsController()
    mappings = mapping_api.HashMapMappingsController()

    @wsme_pecan.wsexpose(service_models.ServiceCollection)
    def get_all(self):
        """Get the service list

        :return: List of every services.
        """
        hashmap = db_api.get_instance()
        service_list = []
        services_uuid_list = hashmap.list_services()
        for service_uuid in services_uuid_list:
            service_db = hashmap.get_service(uuid=service_uuid)
            service_list.append(service_models.Service(
                **service_db.export_model()))
        res = service_models.ServiceCollection(services=service_list)
        return res

    @wsme_pecan.wsexpose(service_models.Service, ck_types.UuidType())
    def get_one(self, service_id):
        """Return a service.

        :param service_id: UUID of the service to filter on.
        """
        hashmap = db_api.get_instance()
        try:
            service_db = hashmap.get_service(uuid=service_id)
            return service_models.Service(**service_db.export_model())
        except db_api.NoSuchService as e:
            pecan.abort(404, six.text_type(e))

    @wsme_pecan.wsexpose(service_models.Service,
                         body=service_models.Service,
                         status_code=201)
    def post(self, service_data):
        """Create hashmap service.

        :param service_data: Informations about the service to create.
        """
        hashmap = db_api.get_instance()
        try:
            service_db = hashmap.create_service(service_data.name)
            pecan.response.location = pecan.request.path_url
            if pecan.response.location[-1] != '/':
                pecan.response.location += '/'
            pecan.response.location += service_db.service_id
            return service_models.Service(
                **service_db.export_model())
        except db_api.ServiceAlreadyExists as e:
            pecan.abort(409, six.text_type(e))

    @wsme_pecan.wsexpose(None, ck_types.UuidType(), status_code=204)
    def delete(self, service_id):
        """Delete the service and all the sub keys recursively.

        :param service_id: UUID of the service to delete.
        """
        hashmap = db_api.get_instance()
        try:
            hashmap.delete_service(uuid=service_id)
        except db_api.NoSuchService as e:
            pecan.abort(404, six.text_type(e))



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Added priority to modules_state.

Revision ID: 385e33fef139
Revises: 2ac2217dcbd9
Create Date: 2015-03-17 17:50:15.229896

"""

# revision identifiers, used by Alembic.
revision = '385e33fef139'
down_revision = '2ac2217dcbd9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        'modules_state',
        sa.Column('priority', sa.Integer(), nullable=True))



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from wsme import types as wtypes

from cloudkitty.api.v1 import types as ck_types


class Group(wtypes.Base):
    """Type describing a group.

    A group is used to divide calculations. It can be used to create a group
    for the instance rating (flavor) and one if we have premium images
    (image_id). So you can take into account multiple parameters during the
    rating.
    """

    group_id = wtypes.wsattr(ck_types.UuidType(), mandatory=False)
    """UUID of the group."""

    name = wtypes.wsattr(wtypes.text, mandatory=True)
    """Name of the group."""

    @classmethod
    def sample(cls):
        sample = cls(group_id='afe898cb-86d8-4557-ad67-f4f01891bbee',
                     name='instance_rating')
        return sample


class GroupCollection(wtypes.Base):
    """Type describing a list of groups.

    """

    groups = [Group]
    """List of groups."""

    @classmethod
    def sample(cls):
        sample = Group.sample()
        return cls(groups=[sample])



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from cloudkitty import transformer


class CeilometerTransformer(transformer.BaseTransformer):
    compute_map = {
        'name': ['display_name'],
        'flavor': ['flavor.name', 'instance_type'],
        'vcpus': ['vcpus'],
        'memory': ['memory_mb'],
        'image_id': ['image.id', 'image_meta.base_image_ref'],
        'availability_zone': [
            'availability_zone',
            'OS-EXT-AZ.availability_zone'],
    }

    volume_map = {
        'volume_id': ['volume_id'],
        'name': ['display_name'],
        'availability_zone': ['availability_zone'],
        'size': ['size'],
    }

    image_map = {
        'container_format': ['container_format'],
        'deleted': ['deleted'],
        'disk_format': ['disk_format'],
        'is_public': ['is_public'],
        'name': ['name'],
        'protected': ['protected'],
        'size': ['size'],
        'status': ['status'],
    }

    network_tap_map = {
        'instance_host': ['instance_host'],
        'mac': ['mac'],
        'host': ['host'],
        'vnic_name': ['vnic_name'],
        'instance_id': ['instance_id'],
    }

    network_floating_map = {
        'status': ['status'],
        'router_id': ['router_id'],
        'floating_network_id': ['floating_network_id'],
        'fixed_ip_address': ['fixed_ip_address'],
        'floating_ip_address': ['floating_ip_address'],
        'port_id': ['port_id'],
    }

    radosgw_usage_map = {}

    metadata_item = 'metadata'

    def _strip_compute(self, data):
        res_data = self.generic_strip('compute', data)
        res_data['instance_id'] = data.resource_id
        res_data['project_id'] = data.project_id
        res_data['user_id'] = data.user_id
        res_data['metadata'] = {}
        for field in data.metadata:
            if field.startswith('user_metadata'):
                res_data['metadata'][field[14:]] = data.metadata[field]
        return res_data

    def _strip_volume(self, data):
        res_data = self.generic_strip('volume', data)
        res_data['user_id'] = data.user_id
        res_data['project_id'] = data.project_id
        return res_data

    def _strip_image(self, data):
        res_data = self.generic_strip('image', data)
        res_data['image_id'] = data.resource_id
        res_data['project_id'] = data.project_id
        res_data['user_id'] = data.user_id
        return res_data

    def _strip_network_tap(self, data):
        res_data = self.generic_strip('network_tap', data)
        res_data['user_id'] = data.user_id
        res_data['project_id'] = data.project_id
        res_data['interface_id'] = data.resource_id
        return res_data

    def _strip_network_floating(self, data):
        res_data = self.generic_strip('network_floating', data)
        res_data['user_id'] = data.user_id
        res_data['project_id'] = data.project_id
        res_data['floatingip_id'] = data.resource_id
        return res_data

    def _strip_radosgw_usage(self, data):
        res_data = self.generic_strip('radosgw_usage_size', data)
        res_data['radosgw_id'] = data.resource_id
        res_data['user_id'] = data.user_id
        res_data['project_id'] = data.project_id
        return res_data

    def get_metadata(self, res_type):
        """Return list of metadata available after transformation for given

        resource type.
        """

        class FakeData(dict):
            """FakeData object."""

            def __getattr__(self, name, default=None):
                try:
                    return super(FakeData, self).__getattr__(self, name)
                except AttributeError:
                    return default or name

        # list of metadata is built by applying the generic strip_resource_data
        # function to a fake data object

        fkdt = FakeData()
        setattr(fkdt, self.metadata_item, FakeData())
        res_data = self.strip_resource_data(res_type, fkdt)
        return res_data.keys()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

import six

from cloudkitty import state
from cloudkitty import utils as ck_utils


@six.add_metaclass(abc.ABCMeta)
class BaseReportWriter(object):
    """Base report writer."""
    report_type = None

    def __init__(self, write_orchestrator, tenant_id, backend, basepath=None):
        self._write_orchestrator = write_orchestrator
        self._backend = backend
        self._tenant_id = tenant_id
        self._sm = state.DBStateManager(self._tenant_id,
                                        self.report_type)
        self._report = None
        self._period = 3600

        self._basepath = basepath

        # State vars
        self.checked_first_line = False
        self.usage_start = None
        self.usage_start_dt = None
        self.usage_end = None
        self.usage_end_dt = None

        # Current total
        self.total = 0

        # Current usage period lines
        self._usage_data = {}

    @abc.abstractmethod
    def _gen_filename(self):
        """Filename generation

        """

    def _open(self):
        filename = self._gen_filename()
        self._report = self._backend(filename, 'wb+')
        self._report.seek(0, 2)

    def _get_report_size(self):
        return self._report.tell()

    @abc.abstractmethod
    def _recover_state(self):
        """Recover state from a last run.

        """

    def _update_state_manager(self):
        self._sm.set_state(self.usage_end)
        metadata = {'total': self.total}
        self._sm.set_metadata(metadata)

    def _get_state_manager_timeframe(self):
        timeframe = self._sm.get_state()
        self.usage_start = timeframe
        self.usage_start_dt = ck_utils.ts2dt(timeframe)
        self.usage_end = timeframe + self._period
        self.usage_end_dt = ck_utils.ts2dt(self.usage_end)
        metadata = self._sm.get_metadata()
        self.total = metadata.get('total', 0)

    def get_timeframe(self, timeframe):
        return self._write_orchestrator.get_timeframe(timeframe)

    @abc.abstractmethod
    def _write_header(self):
        """Write report headers

        """

    @abc.abstractmethod
    def _write_total(self):
        """Write current total

        """

    @abc.abstractmethod
    def _write(self):
        """Write report content

        """

    def _pre_commit(self):
        if self._report is None:
            self._open()
            if not self.checked_first_line:
                if self._get_report_size() == 0:
                    self._write_header()
                else:
                    self._recover_state()
                self.checked_first_line = True
        else:
            self._recover_state()

    def _commit(self):
        self._pre_commit()

        self._write()
        self._update_state_manager()

        self._post_commit()

    def _post_commit(self):
        self._usage_data = {}
        self._write_total()

    def _update(self, data):
        for service in data:
            if service in self._usage_data:
                self._usage_data[service].extend(data[service])
            else:
                self._usage_data[service] = data[service]
            # Update totals
            for entry in data[service]:
                self.total += entry['rating']['price']

    def append(self, data, start, end):
        # FIXME we should use the real time values
        if self.usage_end is not None and start >= self.usage_end:
            self.usage_start = None

        if self.usage_start is None:
            self.usage_start = start
            self.usage_end = start + self._period
            self.usage_start_dt = ck_utils.ts2dt(self.usage_start)
            self.usage_end_dt = ck_utils.ts2dt(self.usage_end)

        self._update(data)

    def commit(self):
        self._commit()

    @abc.abstractmethod
    def _close_file(self):
        """Close report file

        """

    def close(self):
        self._close_file()



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

from oslo_config import cfg
from oslo_db import api as db_api
import six

_BACKEND_MAPPING = {'sqlalchemy': 'cloudkitty.db.sqlalchemy.api'}
IMPL = db_api.DBAPI.from_config(cfg.CONF,
                                backend_mapping=_BACKEND_MAPPING,
                                lazy=True)


def get_instance():
    """Return a DB API instance."""
    return IMPL


@six.add_metaclass(abc.ABCMeta)
class State(object):
    """Base class for state tracking."""

    @abc.abstractmethod
    def get_state(self, name):
        """Retrieve the current state.

        :param name: Name of the state
        :return float: State value
        """

    @abc.abstractmethod
    def set_state(self, name, state):
        """Store the state.

        :param name: Name of the state
        :param state: State value
        """

    @abc.abstractmethod
    def get_metadata(self, name):
        """Retrieve state metadata

        :param name: Name of the state
        :return str: Return a json dict with all metadata attached to this
        state.
        """

    @abc.abstractmethod
    def set_metadata(self, name, metadata):
        """Store the state metadata.

        :param name: Name of the state
        :param metadata: Metadata value
        """


@six.add_metaclass(abc.ABCMeta)
class ModuleInfo(object):
    """Base class for module info management."""

    @abc.abstractmethod
    def get_priority(self, name):
        """Retrieve the module priority.

        :param name: Name of the module
        :return int: Priority of the module
        """

    @abc.abstractmethod
    def set_priority(self, name, priority):
        """Set the module state.

        :param name: Name of the module
        :param priority: New priority of the module
        """

    @abc.abstractmethod
    def get_state(self, name):
        """Retrieve the module state.

        :param name: Name of the module
        :return bool: State of the module
        """

    @abc.abstractmethod
    def set_state(self, name, state):
        """Set the module state.

        :param name: Name of the module
        :param value: State of the module
        """


class NoSuchMapping(Exception):
    """Raised when the mapping doesn't exist."""

    def __init__(self, service):
        super(NoSuchMapping, self).__init__(
            "No mapping for service: %s" % service)
        self.service = service


@six.add_metaclass(abc.ABCMeta)
class ServiceToCollectorMapping(object):
    """Base class for service to collector mapping."""

    @abc.abstractmethod
    def get_mapping(self, service):
        """Get a mapping.

        :return mapping: service to collector object.
        """

    @abc.abstractmethod
    def set_mapping(self, service, collector):
        """Set a mapping.

        :param service: Service to work on.
        :param collector: Collector to prioritize.
        :return mapping: Service to Collector object.
        """

    @abc.abstractmethod
    def list_services(self, collector=None):
        """Retrieve the list of every services mapped.

        :param collector: Filter on a collector name.
        :return list(str): List of services' name.
        """
    @abc.abstractmethod
    def list_mappings(self, collector=None):
        """Retrieve the list of every mappings.

        :param collector: Filter on a collector's name.
        :return [tuple(str, str)]: List of mappings.
        """

    @abc.abstractmethod
    def delete_mapping(self, service):
        """Remove a mapping.

        """



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal
import json

from cloudkitty.db import api


class StateManager(object):
    def __init__(self, state_backend, state_basepath, user_id, report_type,
                 distributed=False):
        self._backend = state_backend
        self._basepath = state_basepath
        self._uid = user_id
        self._type = report_type
        self._distributed = distributed

        # States
        self._ts = None
        self._metadata = {}

        # Load states
        self._load()

    def _gen_filename(self):
        # FIXME(sheeprine): Basepath can't be enforced at the moment
        filename = '{0}_{1}.state'.format(self._type,
                                          self._uid)
        return filename

    def _open(self, mode='rb'):
        filename = self._gen_filename()
        state_file = self._backend(filename, mode)
        return state_file

    def _load(self):
        try:
            state_file = self._open()
            raw_data = state_file.read()
            if raw_data:
                state_data = json.loads(raw_data)
                self._ts = state_data['timestamp']
                self._metadata = state_data['metadata']
            state_file.close()
        except IOError:
            pass

    def _update(self):
        state_file = self._open('wb')
        state_data = {'timestamp': self._ts,
                      'metadata': self._metadata}
        state_file.write(json.dumps(state_data))
        state_file.close()

    def set_state(self, timestamp):
        """Set the current state's timestamp."""
        if self._distributed:
            self._load()
        self._ts = timestamp
        self._update()

    def get_state(self):
        """Get the state timestamp."""
        if self._distributed:
            self._load()
        return self._ts

    def set_metadata(self, metadata):
        """Set metadata attached to the state."""
        if self._distributed:
            self._load()
        self._metadata = metadata
        self._update()

    def get_metadata(self):
        """Get metadata attached to the state."""
        if self._distributed:
            self._load()
        return self._metadata


class DecimalJSONEncoder(json.JSONEncoder):
    """Wrapper class to handle decimal.Decimal objects in json.dumps()."""
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        return super(DecimalJSONEncoder, self).default(obj)


class DBStateManager(object):
    def __init__(self, user_id, report_type, distributed=False):
        self._state_name = self._gen_name(report_type, user_id)
        self._distributed = distributed
        self._db = api.get_instance().get_state()

    def _gen_name(self, state_type, uid):
        name = '{0}_{1}'.format(state_type, uid)
        return name

    def get_state(self):
        """Get the state timestamp."""

        return self._db.get_state(self._state_name)

    def set_state(self, timestamp):
        """Set the current state's timestamp."""

        self._db.set_state(self._state_name, timestamp)

    def get_metadata(self):
        """Get metadata attached to the state."""

        data = self._db.get_metadata(self._state_name)
        if data:
            return json.loads(data)

    def set_metadata(self, metadata):
        """Set metadata attached to the state."""

        self._db.set_metadata(self._state_name,
                              json.dumps(metadata, cls=DecimalJSONEncoder))



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from cloudkitty.common.db.alembic import env  # noqa
from cloudkitty.rating.hash.db.sqlalchemy import models

target_metadata = models.Base.metadata
version_table = 'hashmap_alembic'


env.run_migrations_online(target_metadata, version_table)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Aaron.Ding(dinghh@awcloud.com)
#
import datetime

from wsme import types as wtypes


class SummaryModel(wtypes.Base):
    """Type describing a report summary info."""

    begin = datetime.datetime
    """Begin date for the sample."""

    end = datetime.datetime
    """End date for the sample."""

    tenant_id = wtypes.text
    """Tenant owner of the sample."""

    res_type = wtypes.text
    """Resource type of the sample."""

    rate = wtypes.text
    """summary rate of the sample"""

    def __init__(self, begin=None, end=None, tenant_id=None,
                 res_type=None, rate=None):
        self.begin = begin
        self.end = end
        self.tenant_id = tenant_id if tenant_id else "ALL"
        self.res_type = res_type if res_type else "ALL"
        # TODO(Aaron): Need optimize, control precision with decimal
        self.rate = str(float('%0.5f' % rate)) if rate else "0"

    def to_json(self):
        return {'begin': self.begin,
                'end': self.end,
                'tenant_id': self.tenant_id,
                'res_type': self.res_type,
                'rate': self.rate}

    @classmethod
    def sample(cls):
        sample = cls(tenant_id='69d12143688f413cbf5c3cfe03ed0a12',
                     begin=datetime.datetime(2015, 4, 22, 7),
                     end=datetime.datetime(2015, 4, 22, 8),
                     res_type='compute',
                     rate="1")
        return sample


class SummaryCollectionModel(wtypes.Base):
    """A list of report summary."""

    summary = [SummaryModel]

    @classmethod
    def sample(cls):
        sample = SummaryModel.sample()
        return cls(summary=[sample])



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import abc

import six


@six.add_metaclass(abc.ABCMeta)
class BaseIOBackend(object):
    def __init__(self, path):
        self.open(path)

    @abc.abstractmethod
    def open(self, path):
        """Open the connection/file on the backend.

        """

    @abc.abstractmethod
    def tell(self):
        """Current position on the backend.

        """

    @abc.abstractmethod
    def seek(self, offset, from_what=0):
        # 0 beg, 1 cur, 2 end
        """Change position in the backend.

        """

    @abc.abstractmethod
    def flush(self):
        """Force write informations on the backend.

        """

    @abc.abstractmethod
    def write(self, data):
        """Writer data on the backend.

        :param data: Data to be written on the backend.
        """

    @abc.abstractmethod
    def read(self):
        """Read data from the backend.

        :return str: Data read from the backend.
        """

    @abc.abstractmethod
    def close(self):
        """Close the connection/file on the backend.

        """



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import copy
import decimal

import mock
from oslo_utils import uuidutils

from cloudkitty.rating import hash
from cloudkitty.rating.hash.db import api
from cloudkitty import tests

TEST_TS = 1388577600
FAKE_UUID = '6c1b8a30-797f-4b7e-ad66-9879b79059fb'
CK_RESOURCES_DATA = [{
    "period": {
        "begin": "2014-10-01T00:00:00",
        "end": "2014-10-01T01:00:00"},
    "usage": {
        "compute": [
            {
                "desc": {
                    "availability_zone": "nova",
                    "flavor": "m1.nano",
                    "image_id": "f5600101-8fa2-4864-899e-ebcb7ed6b568",
                    "memory": "64",
                    "metadata": {
                        "farm": "prod"},
                    "name": "prod1",
                    "project_id": "f266f30b11f246b589fd266f85eeec39",
                    "user_id": "55b3379b949243009ee96972fbf51ed1",
                    "vcpus": "1"},
                "vol": {
                    "qty": 1,
                    "unit": "instance"}
            },
            {
                "desc": {
                    "availability_zone": "nova",
                    "flavor": "m1.tiny",
                    "image_id": "a41fba37-2429-4f15-aa00-b5bc4bf557bf",
                    "memory": "512",
                    "metadata": {
                        "farm": "dev"},
                    "name": "dev1",
                    "project_id": "f266f30b11f246b589fd266f85eeec39",
                    "user_id": "55b3379b949243009ee96972fbf51ed1",
                    "vcpus": "1"},
                "vol": {
                    "qty": 2,
                    "unit": "instance"}},
            {
                "desc": {
                    "availability_zone": "nova",
                    "flavor": "m1.nano",
                    "image_id": "a41fba37-2429-4f15-aa00-b5bc4bf557bf",
                    "memory": "64",
                    "metadata": {
                        "farm": "dev"},
                    "name": "dev2",
                    "project_id": "f266f30b11f246b589fd266f85eeec39",
                    "user_id": "55b3379b949243009ee96972fbf51ed1",
                    "vcpus": "1"},
                "vol": {
                    "qty": 1,
                    "unit": "instance"}}]}}]


class HashMapRatingTest(tests.TestCase):
    def setUp(self):
        super(HashMapRatingTest, self).setUp()
        self._tenant_id = 'f266f30b11f246b589fd266f85eeec39'
        self._db_api = hash.HashMap.db_api
        self._db_api.get_migration().upgrade('head')
        self._hash = hash.HashMap(self._tenant_id)

    # Group tests
    @mock.patch.object(uuidutils, 'generate_uuid',
                       return_value=FAKE_UUID)
    def test_create_group(self, patch_generate_uuid):
        self._db_api.create_group('test_group')
        groups = self._db_api.list_groups()
        self.assertEqual([FAKE_UUID], groups)
        patch_generate_uuid.assert_called_once_with()

    def test_create_duplicate_group(self):
        self._db_api.create_group('test_group')
        self.assertRaises(api.GroupAlreadyExists,
                          self._db_api.create_group,
                          'test_group')

    def test_delete_group(self):
        group_db = self._db_api.create_group('test_group')
        self._db_api.delete_group(group_db.group_id)
        groups = self._db_api.list_groups()
        self.assertEqual([], groups)

    def test_delete_unknown_group(self):
        self.assertRaises(api.NoSuchGroup,
                          self._db_api.delete_group,
                          uuidutils.generate_uuid())

    def test_recursive_delete_group(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        group_db = self._db_api.create_group('test_group')
        self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            group_id=group_db.group_id)
        self._db_api.delete_group(group_db.group_id)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id)
        self.assertEqual([], mappings)
        groups = self._db_api.list_groups()
        self.assertEqual([], groups)

    def test_non_recursive_delete_group(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        group_db = self._db_api.create_group('test_group')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            group_id=group_db.group_id)
        self._db_api.delete_group(group_db.group_id, False)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id)
        self.assertEqual([mapping_db.mapping_id], mappings)
        groups = self._db_api.list_groups()
        self.assertEqual([], groups)
        new_mapping_db = self._db_api.get_mapping(mapping_db.mapping_id)
        self.assertIsNone(new_mapping_db.group_id)

    def test_list_mappings_from_only_group(self):
        service_db = self._db_api.create_service('compute')
        group_db = self._db_api.create_group('test_group')
        mapping_tiny = self._db_api.create_mapping(
            cost='1.337',
            map_type='flat',
            service_id=service_db.service_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            cost='42',
            map_type='flat',
            service_id=service_db.service_id)
        mappings = self._db_api.list_mappings(group_uuid=group_db.group_id)
        self.assertEqual([mapping_tiny.mapping_id],
                         mappings)

    def test_list_mappings_from_group(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        group_db = self._db_api.create_group('test_group')
        mapping_tiny = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            group_id=group_db.group_id)
        mapping_small = self._db_api.create_mapping(
            value='m1.small',
            cost='3.1337',
            map_type='flat',
            field_id=field_db.field_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            value='m1.large',
            cost='42',
            map_type='flat',
            field_id=field_db.field_id)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id,
                                              group_uuid=group_db.group_id)
        self.assertEqual([mapping_tiny.mapping_id,
                          mapping_small.mapping_id],
                         mappings)

    def test_list_mappings_without_group(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        group_db = self._db_api.create_group('test_group')
        self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            value='m1.small',
            cost='3.1337',
            map_type='flat',
            field_id=field_db.field_id,
            group_id=group_db.group_id)
        mapping_no_group = self._db_api.create_mapping(
            value='m1.large',
            cost='42',
            map_type='flat',
            field_id=field_db.field_id)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id,
                                              no_group=True)
        self.assertEqual([mapping_no_group.mapping_id],
                         mappings)

    # Service tests
    @mock.patch.object(uuidutils, 'generate_uuid',
                       return_value=FAKE_UUID)
    def test_create_service(self, patch_generate_uuid):
        self._db_api.create_service('compute')
        services = self._db_api.list_services()
        self.assertEqual([FAKE_UUID], services)
        patch_generate_uuid.assert_called_once_with()

    def test_create_duplicate_service(self):
        self._db_api.create_service('compute')
        self.assertRaises(api.ServiceAlreadyExists,
                          self._db_api.create_service,
                          'compute')

    def test_delete_service_by_name(self):
        self._db_api.create_service('compute')
        self._db_api.delete_service('compute')
        services = self._db_api.list_services()
        self.assertEqual([], services)

    def test_delete_service_by_uuid(self):
        service_db = self._db_api.create_service('compute')
        self._db_api.delete_service(uuid=service_db.service_id)
        services = self._db_api.list_services()
        self.assertEqual([], services)

    def test_delete_unknown_service_by_name(self):
        self.assertRaises(api.NoSuchService,
                          self._db_api.delete_service,
                          'dummy')

    def test_delete_unknown_service_by_uuid(self):
        self.assertRaises(
            api.NoSuchService,
            self._db_api.delete_service,
            uuid='6e8de9fc-ee17-4b60-b81a-c9320e994e76')

    # Field tests
    def test_create_field_in_existing_service(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        fields = self._db_api.list_fields(service_db.service_id)
        self.assertEqual([field_db.field_id], fields)

    def test_create_duplicate_field(self):
        service_db = self._db_api.create_service('compute')
        self._db_api.create_field(service_db.service_id,
                                  'flavor')
        self.assertRaises(api.FieldAlreadyExists,
                          self._db_api.create_field,
                          service_db.service_id,
                          'flavor')

    def test_delete_field(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id, 'flavor')
        self._db_api.delete_field(field_db.field_id)
        services = self._db_api.list_services()
        self.assertEqual([service_db.service_id], services)
        fields = self._db_api.list_fields(service_db.service_id)
        self.assertEqual([], fields)

    def test_delete_unknown_field(self):
        self.assertRaises(api.NoSuchField,
                          self._db_api.delete_field,
                          uuidutils.generate_uuid())

    def test_recursive_delete_field_from_service(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        self._db_api.delete_service(uuid=service_db.service_id)
        self.assertRaises(api.NoSuchField,
                          self._db_api.get_field,
                          field_db.field_id)

    # Mapping tests
    def test_create_mapping(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id)
        self.assertEqual([mapping_db.mapping_id], mappings)

    def test_get_mapping(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        mapping = self._db_api.get_mapping(mapping_db.mapping_id)
        self.assertEqual('flat', mapping.map_type)
        self.assertEqual('m1.tiny', mapping.value)
        self.assertEqual(decimal.Decimal('1.337'), mapping.cost)
        self.assertEqual(field_db.id, mapping.field_id)

    def test_list_mappings_from_services(self):
        service_db = self._db_api.create_service('compute')
        mapping_db = self._db_api.create_mapping(
            cost='1.337',
            map_type='flat',
            service_id=service_db.service_id)
        mappings = self._db_api.list_mappings(
            service_uuid=service_db.service_id)
        self.assertEqual([mapping_db.mapping_id], mappings)

    def test_list_mappings_from_fields(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        mappings = self._db_api.list_mappings(
            field_uuid=field_db.field_id)
        self.assertEqual([mapping_db.mapping_id], mappings)

    def test_create_mapping_with_incorrect_type(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        self.assertRaises(api.NoSuchType,
                          self._db_api.create_mapping,
                          value='m1.tiny',
                          cost='1.337',
                          map_type='invalid',
                          field_id=field_db.field_id)

    def test_create_mapping_with_two_parents(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        self.assertRaises(api.ClientHashMapError,
                          self._db_api.create_mapping,
                          value='m1.tiny',
                          cost='1.337',
                          map_type='flat',
                          service_id=service_db.service_id,
                          field_id=field_db.field_id)

    def test_update_mapping(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        new_mapping_db = self._db_api.update_mapping(
            uuid=mapping_db.mapping_id,
            value='42',
            map_type='rate')
        self.assertEqual('42', new_mapping_db.value)
        self.assertEqual('rate', new_mapping_db.map_type)

    def test_update_mapping_inside_group(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        group_db = self._db_api.create_group('test_group')
        new_mapping_db = self._db_api.update_mapping(
            mapping_db.mapping_id,
            value='42',
            map_type='rate',
            group_id=group_db.group_id)
        self.assertEqual('42', new_mapping_db.value)
        self.assertEqual('rate', new_mapping_db.map_type)
        self.assertEqual(group_db.id, new_mapping_db.group_id)

    def test_delete_mapping(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        self._db_api.delete_mapping(mapping_db.mapping_id)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id)
        self.assertEqual([], mappings)

    def test_create_per_tenant_mapping(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            tenant_id=self._tenant_id)
        mappings = self._db_api.list_mappings(field_uuid=field_db.field_id)
        self.assertEqual(
            self._tenant_id,
            mapping_db.tenant_id)
        self.assertEqual([mapping_db.mapping_id], mappings)

    def test_list_mappings_filtering_on_tenant(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            tenant_id=self._tenant_id)
        self._db_api.create_mapping(
            value='m1.small',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        mappings = self._db_api.list_mappings(
            field_uuid=field_db.field_id,
            tenant_uuid=self._tenant_id)
        self.assertEqual([mapping_db.mapping_id], mappings)

    def test_list_mappings_filtering_on_no_tenant(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'flavor')
        mapping_db = self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id)
        self._db_api.create_mapping(
            value='m1.small',
            cost='1.337',
            map_type='flat',
            field_id=field_db.field_id,
            tenant_id=self._tenant_id)
        mappings = self._db_api.list_mappings(
            field_uuid=field_db.field_id,
            tenant_uuid=None)
        self.assertEqual([mapping_db.mapping_id], mappings)

    # Threshold tests
    def test_create_threshold(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        thresholds = self._db_api.list_thresholds(field_uuid=field_db.field_id)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    def test_get_threshold(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='rate',
            field_id=field_db.field_id)
        threshold = self._db_api.get_threshold(threshold_db.threshold_id)
        self.assertEqual('rate', threshold.map_type)
        self.assertEqual(decimal.Decimal('64'), threshold.level)
        self.assertEqual(decimal.Decimal('0.1337'), threshold.cost)
        self.assertEqual(field_db.id, threshold.field_id)

    def test_list_thresholds_from_only_group(self):
        service_db = self._db_api.create_service('compute')
        group_db = self._db_api.create_group('test_group')
        threshold_db = self._db_api.create_threshold(
            level=10,
            cost='1.337',
            map_type='flat',
            service_id=service_db.service_id,
            group_id=group_db.group_id)
        thresholds = self._db_api.list_thresholds(
            group_uuid=group_db.group_id)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    def test_list_thresholds_from_services(self):
        service_db = self._db_api.create_service('compute')
        threshold_db = self._db_api.create_threshold(
            level=10,
            cost='1.337',
            map_type='flat',
            service_id=service_db.service_id)
        thresholds = self._db_api.list_thresholds(
            service_uuid=service_db.service_id)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    def test_list_thresholds_from_fields(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        thresholds = self._db_api.list_thresholds(field_uuid=field_db.field_id)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    def test_create_threshold_with_incorrect_type(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        self.assertRaises(
            api.NoSuchType,
            self._db_api.create_threshold,
            level='64',
            cost='0.1337',
            map_type='invalid',
            field_id=field_db.field_id)

    def test_create_threshold_with_two_parents(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        self.assertRaises(
            api.ClientHashMapError,
            self._db_api.create_threshold,
            level='64',
            cost='0.1337',
            map_type='flat',
            service_id=service_db.service_id,
            field_id=field_db.field_id)

    def test_update_threshold(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        new_threshold_db = self._db_api.update_threshold(
            uuid=threshold_db.threshold_id,
            level='128',
            map_type='rate')
        self.assertEqual('128', new_threshold_db.level)
        self.assertEqual('rate', new_threshold_db.map_type)

    def test_update_threshold_inside_group(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        group_db = self._db_api.create_group('test_group')
        new_threshold_db = self._db_api.update_threshold(
            threshold_db.threshold_id,
            group_id=group_db.group_id)
        self.assertEqual(group_db.id, new_threshold_db.group_id)

    def test_delete_threshold(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        self._db_api.delete_threshold(threshold_db.threshold_id)
        thresholds = self._db_api.list_thresholds(field_uuid=field_db.field_id)
        self.assertEqual([], thresholds)

    def test_create_per_tenant_threshold(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id,
            tenant_id=self._tenant_id)
        thresholds = self._db_api.list_thresholds(field_uuid=field_db.field_id)
        self.assertEqual(
            self._tenant_id,
            threshold_db.tenant_id)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    def test_list_thresholds_filtering_on_tenant(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id,
            tenant_id=self._tenant_id)
        self._db_api.create_threshold(
            level='128',
            cost='0.2',
            map_type='flat',
            field_id=field_db.field_id)
        thresholds = self._db_api.list_thresholds(
            field_uuid=field_db.field_id,
            tenant_uuid=self._tenant_id)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    def test_list_thresholds_filtering_on_no_tenant(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(
            service_db.service_id,
            'memory')
        threshold_db = self._db_api.create_threshold(
            level='64',
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        self._db_api.create_threshold(
            level='128',
            cost='0.2',
            map_type='flat',
            field_id=field_db.field_id,
            tenant_id=self._tenant_id)
        thresholds = self._db_api.list_thresholds(
            field_uuid=field_db.field_id,
            tenant_uuid=None)
        self.assertEqual([threshold_db.threshold_id], thresholds)

    # Processing tests
    def _generate_hashmap_rules(self):
        mapping_list = []
        threshold_list = []
        service_db = self._db_api.create_service('compute')
        flavor_field = self._db_api.create_field(service_db.service_id,
                                                 'flavor')
        memory_field = self._db_api.create_field(service_db.service_id,
                                                 'memory')
        group_db = self._db_api.create_group('test_group')
        mapping_list.append(
            self._db_api.create_mapping(
                cost='1.42',
                map_type='rate',
                service_id=service_db.service_id))
        mapping_list.append(
            self._db_api.create_mapping(
                value='m1.tiny',
                cost='1.337',
                map_type='flat',
                field_id=flavor_field.field_id))
        mapping_list.append(
            self._db_api.create_mapping(
                value='m1.large',
                cost='13.37',
                map_type='rate',
                field_id=flavor_field.field_id,
                group_id=group_db.group_id))
        # Per tenant override
        mapping_list.append(
            self._db_api.create_mapping(
                value='m1.tiny',
                cost='2',
                map_type='flat',
                field_id=flavor_field.field_id,
                tenant_id=self._tenant_id))
        threshold_list.append(
            self._db_api.create_threshold(
                level='64',
                cost='0.02',
                map_type='flat',
                field_id=memory_field.field_id,
                group_id=group_db.group_id))
        threshold_list.append(
            self._db_api.create_threshold(
                level='128',
                cost='0.03',
                map_type='flat',
                field_id=memory_field.field_id,
                group_id=group_db.group_id))
        threshold_list.append(
            self._db_api.create_threshold(
                level='64',
                cost='0.03',
                map_type='flat',
                field_id=memory_field.field_id,
                group_id=group_db.group_id,
                tenant_id=self._tenant_id))
        return ([mapping.mapping_id for mapping in mapping_list],
                [threshold.threshold_id for threshold in threshold_list])

    def test_load_rates(self):
        self._generate_hashmap_rules()
        self._hash.reload_config()
        expect = {
            'compute': {
                'fields': {
                    'flavor': {
                        'mappings': {
                            '_DEFAULT_': {
                                'm1.tiny': {
                                    'cost': decimal.Decimal('2'),
                                    'type': 'flat'}},
                            'test_group': {
                                'm1.large': {
                                    'cost': decimal.Decimal('13.37'),
                                    'type': 'rate'}}},
                        'thresholds': {}},
                    'memory': {
                        'mappings': {},
                        'thresholds': {
                            'test_group': {
                                64: {
                                    'cost': decimal.Decimal('0.03'),
                                    'type': 'flat'},
                                128: {
                                    'cost': decimal.Decimal('0.03'),
                                    'type': 'flat'}}}}},
                'mappings': {
                    '_DEFAULT_': {
                        'cost': decimal.Decimal('1.42'),
                        'type': 'rate'}},
                'thresholds': {}}}
        self.assertEqual(expect,
                         self._hash._entries)

    def test_load_mappings(self):
        mapping_list = []
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        group_db = self._db_api.create_group('test_group')
        mapping_list.append(
            self._db_api.create_mapping(
                value='m1.tiny',
                cost='1.337',
                map_type='flat',
                field_id=field_db.field_id))
        mapping_list.append(
            self._db_api.create_mapping(
                value='m1.large',
                cost='13.37',
                map_type='rate',
                field_id=field_db.field_id,
                group_id=group_db.group_id))
        mappings_uuid = [mapping.mapping_id for mapping in mapping_list]
        result = self._hash._load_mappings(mappings_uuid)
        expected_result = {
            '_DEFAULT_': {
                'm1.tiny': {
                    'cost': decimal.Decimal('1.337'),
                    'type': 'flat'}},
            'test_group': {
                'm1.large': {
                    'cost': decimal.Decimal('13.37'),
                    'type': 'rate'}}}
        self.assertEqual(expected_result, result)

    def test_load_thresholds(self):
        threshold_list = []
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'flavor')
        group_db = self._db_api.create_group('test_group')
        threshold_list.append(
            self._db_api.create_threshold(
                level='1000',
                cost='3.1337',
                map_type='flat',
                field_id=field_db.field_id,
                group_id=group_db.group_id))
        thresholds_uuid = [threshold.threshold_id
                           for threshold in threshold_list]
        result = self._hash._load_thresholds(thresholds_uuid)
        expected_result = {
            'test_group': {
                1000: {
                    'cost': decimal.Decimal('3.1337'),
                    'type': 'flat'}}}
        self.assertEqual(expected_result, result)

    def test_process_services(self):
        service_db = self._db_api.create_service('compute')
        group_db = self._db_api.create_group('test_group')
        self._db_api.create_mapping(
            cost='1.337',
            map_type='flat',
            service_id=service_db.service_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            cost='1.42',
            map_type='flat',
            service_id=service_db.service_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        for cur_data in actual_data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._hash._res = {}
                    self._hash.process_services(service_name, item)
                    self._hash.add_rating_informations(item)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('2.757')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('5.514')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('2.757')}
        self.assertEqual(expected_data, actual_data)

    def test_process_fields(self):
        service_db = self._db_api.create_service('compute')
        flavor_field = self._db_api.create_field(service_db.service_id,
                                                 'flavor')
        image_field = self._db_api.create_field(service_db.service_id,
                                                'image_id')
        group_db = self._db_api.create_group('test_group')
        self._db_api.create_mapping(
            value='m1.nano',
            cost='1.337',
            map_type='flat',
            field_id=flavor_field.field_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            value='a41fba37-2429-4f15-aa00-b5bc4bf557bf',
            cost='1.10',
            map_type='rate',
            field_id=image_field.field_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.42',
            map_type='flat',
            field_id=flavor_field.field_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        for cur_data in actual_data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._hash._res = {}
                    self._hash.process_fields(service_name, item)
                    self._hash.add_rating_informations(item)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('1.337')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('2.84')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('1.47070')}
        self.assertEqual(expected_data, actual_data)

    def test_process_fields_no_match(self):
        service_db = self._db_api.create_service('compute')
        flavor_field = self._db_api.create_field(service_db.service_id,
                                                 'flavor')
        self._db_api.create_mapping(
            value='non-existent',
            cost='1.337',
            map_type='flat',
            field_id=flavor_field.field_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        for cur_data in actual_data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._hash._res = {}
                    self._hash.process_fields(service_name, item)
                    self._hash.add_rating_informations(item)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('0')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('0')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('0')}
        self.assertEqual(expected_data, actual_data)

    def test_process_field_threshold(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'memory')
        self._db_api.create_threshold(
            level=64,
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        self._db_api.create_threshold(
            level=128,
            cost='0.2',
            map_type='flat',
            field_id=field_db.field_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        for cur_data in actual_data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._hash._res = {}
                    self._hash.process_fields(service_name, item)
                    self._hash.add_rating_informations(item)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('0.1337')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('0.4')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('0.1337')}
        self.assertEqual(expected_data, actual_data)

    def test_process_field_threshold_no_match(self):
        service_db = self._db_api.create_service('compute')
        field_db = self._db_api.create_field(service_db.service_id,
                                             'memory')
        self._db_api.create_threshold(
            level=10240,
            cost='0.1337',
            map_type='flat',
            field_id=field_db.field_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        for cur_data in actual_data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._hash._res = {}
                    self._hash.process_fields(service_name, item)
                    self._hash.add_rating_informations(item)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('0')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('0')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('0')}
        self.assertEqual(expected_data, actual_data)

    def test_process_service_threshold(self):
        service_db = self._db_api.create_service('compute')
        self._db_api.create_threshold(
            level=1,
            cost='0.1',
            map_type='flat',
            service_id=service_db.service_id)
        self._db_api.create_threshold(
            level=2,
            cost='0.15',
            map_type='flat',
            service_id=service_db.service_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        for cur_data in actual_data:
            cur_usage = cur_data['usage']
            for service_name, service_data in cur_usage.items():
                for item in service_data:
                    self._hash._res = {}
                    self._hash.process_services(service_name, item)
                    self._hash.add_rating_informations(item)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('0.1')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('0.15')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('0.1')}
        self.assertEqual(expected_data, actual_data)

    def test_update_result_flat(self):
        self._hash.update_result(
            'test_group',
            'flat',
            1)
        self.assertEqual(1, self._hash._res['test_group']['flat'])
        self._hash.update_result(
            'test_group',
            'flat',
            0.5)
        self.assertEqual(1, self._hash._res['test_group']['flat'])
        self._hash.update_result(
            'test_group',
            'flat',
            1.5)
        self.assertEqual(1.5, self._hash._res['test_group']['flat'])

    def test_update_result_rate(self):
        self._hash.update_result(
            'test_group',
            'rate',
            0.5)
        self.assertEqual(0.5, self._hash._res['test_group']['rate'])
        self._hash.update_result(
            'test_group',
            'rate',
            0.5)
        self.assertEqual(0.25, self._hash._res['test_group']['rate'])
        self._hash.update_result(
            'test_group',
            'rate',
            1)
        self.assertEqual(0.25, self._hash._res['test_group']['rate'])

    def test_update_result_threshold(self):
        self._hash.update_result(
            'test_group',
            'flat',
            0.01,
            0,
            True)
        self.assertEqual({'level': 0,
                          'cost': 0.01,
                          'scope': 'field',
                          'type': 'flat'},
                         self._hash._res['test_group']['threshold'])
        self._hash.update_result(
            'test_group',
            'flat',
            1,
            10,
            True)
        self.assertEqual({'level': 10,
                          'cost': 1,
                          'scope': 'field',
                          'type': 'flat'},
                         self._hash._res['test_group']['threshold'])
        self._hash.update_result(
            'test_group',
            'flat',
            1.1,
            15,
            True)
        self.assertEqual({'level': 15,
                          'cost': 1.1,
                          'scope': 'field',
                          'type': 'flat'},
                         self._hash._res['test_group']['threshold'])
        self._hash.update_result(
            'test_group',
            'threshold',
            2.2,
            10,
            True)
        self.assertEqual({'level': 15,
                          'cost': 1.1,
                          'scope': 'field',
                          'type': 'flat'},
                         self._hash._res['test_group']['threshold'])

    def test_process_rating(self):
        service_db = self._db_api.create_service('compute')
        flavor_db = self._db_api.create_field(service_db.service_id,
                                              'flavor')
        group_db = self._db_api.create_group('test_group')
        self._db_api.create_mapping(
            cost='1.00',
            map_type='flat',
            service_id=service_db.service_id)
        self._db_api.create_mapping(
            value='m1.nano',
            cost='1.337',
            map_type='flat',
            field_id=flavor_db.field_id,
            group_id=group_db.group_id)
        self._db_api.create_mapping(
            value='m1.tiny',
            cost='1.42',
            map_type='flat',
            field_id=flavor_db.field_id,
            group_id=group_db.group_id)
        image_db = self._db_api.create_field(service_db.service_id,
                                             'image_id')
        self._db_api.create_mapping(
            value='a41fba37-2429-4f15-aa00-b5bc4bf557bf',
            cost='1.10',
            map_type='rate',
            field_id=image_db.field_id,
            group_id=group_db.group_id)
        memory_db = self._db_api.create_field(service_db.service_id,
                                              'memory')
        self._db_api.create_threshold(
            level=64,
            cost='0.15',
            map_type='flat',
            field_id=memory_db.field_id,
            group_id=group_db.group_id)
        self._db_api.create_threshold(
            level=128,
            cost='0.2',
            map_type='flat',
            field_id=memory_db.field_id,
            group_id=group_db.group_id)
        self._hash.reload_config()
        actual_data = copy.deepcopy(CK_RESOURCES_DATA)
        expected_data = copy.deepcopy(CK_RESOURCES_DATA)
        compute_list = expected_data[0]['usage']['compute']
        compute_list[0]['rating'] = {'price': decimal.Decimal('2.487')}
        compute_list[1]['rating'] = {'price': decimal.Decimal('5.564')}
        compute_list[2]['rating'] = {'price': decimal.Decimal('2.6357')}
        self._hash.process(actual_data)
        self.assertEqual(expected_data, actual_data)



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from cloudkitty import rating
from cloudkitty.rating.pyscripts.controllers import root as root_api
from cloudkitty.rating.pyscripts.db import api as pyscripts_db_api


class PyScripts(rating.RatingProcessorBase):
    """PyScripts rating module.

    PyScripts is a module made to execute custom made python scripts to create
    rating policies.
    """

    module_name = 'pyscripts'
    description = 'PyScripts rating module.'
    hot_config = True
    config_controller = root_api.PyScriptsConfigController

    db_api = pyscripts_db_api.get_instance()

    def __init__(self, tenant_id=None):
        self._scripts = {}
        self.load_scripts_in_memory()
        super(PyScripts, self).__init__(tenant_id)

    def load_scripts_in_memory(self):
        db = pyscripts_db_api.get_instance()
        scripts_uuid_list = db.list_scripts()
        # Purge old entries
        scripts_to_purge = []
        for script_uuid in self._scripts.keys():
            if script_uuid not in scripts_uuid_list:
                scripts_to_purge.append(script_uuid)
        for script_uuid in scripts_to_purge:
            del self._scripts[script_uuid]
        # Load or update script
        for script_uuid in scripts_uuid_list:
            script_db = db.get_script(uuid=script_uuid)
            name = script_db.name
            checksum = script_db.checksum
            if name not in self._scripts:
                self._scripts[script_uuid] = {}
            script = self._scripts[script_uuid]
            # NOTE(sheeprine): We're doing this the easy way, we might want to
            # store the context and call functions in future
            if script.get(checksum, '') != checksum:
                code = compile(
                    script_db.data,
                    '<PyScripts: {name}>'.format(name=name),
                    'exec')
                script.update({
                    'name': name,
                    'code': code,
                    'checksum': checksum})

    def reload_config(self):
        """Reload the module's configuration.

        """
        self.load_scripts_in_memory()

    def start_script(self, code, data):
        context = {'data': data}
        exec(code, context)
        return data

    def process(self, data):
        for script in self._scripts.values():
            self.start_script(script['code'], data)
        return data



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import datetime
import decimal

from wsme import types as wtypes

from cloudkitty.api.v1.datamodels import rating as rating_resources


class RatedResource(rating_resources.CloudkittyResource):
    """Represents a rated CloudKitty resource."""

    rating = decimal.Decimal

    def to_json(self):
        res_dict = super(RatedResource, self).to_json()
        res_dict['rating'] = self.rating
        return res_dict

    @classmethod
    def sample(cls):
        sample = cls(volume=decimal.Decimal('1.0'),
                     service='compute',
                     rating=decimal.Decimal('1.0'),
                     desc={'flavor': 'm1.tiny', 'vcpus': '1'})
        return sample


class DataFrame(wtypes.Base):
    """Type describing a stored data frame."""

    begin = datetime.datetime
    """Begin date for the sample."""

    end = datetime.datetime
    """End date for the sample."""

    tenant_id = wtypes.text
    """Tenant owner of the sample."""

    resources = [RatedResource]
    """A resource list."""

    def to_json(self):
        return {'begin': self.begin,
                'end': self.end,
                'tenant_id': self.tenant_id,
                'resources': self.resources}

    @classmethod
    def sample(cls):
        res_sample = RatedResource.sample()
        sample = cls(tenant_id='69d12143688f413cbf5c3cfe03ed0a12',
                     begin=datetime.datetime(2015, 4, 22, 7),
                     end=datetime.datetime(2015, 4, 22, 8),
                     resources=[res_sample])
        return sample


class DataFrameCollection(wtypes.Base):
    """A list of stored data frames."""

    dataframes = [DataFrame]

    @classmethod
    def sample(cls):
        sample = DataFrame.sample()
        return cls(dataframes=[sample])



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal

from oslo_log import log

from cloudkitty.storage.gnocchi_hybrid import migration
from cloudkitty.storage.gnocchi_hybrid import models
from cloudkitty.storage import sqlalchemy as sql_storage

LOG = log.getLogger(__name__)


class GnocchiHybridStorage(sql_storage.SQLAlchemyStorage):
    """Gnocchi Hybrid Storage Backend

    Driver used to add support for gnocchi until the creation of custom
    resources is supported in gnocchi.
    """
    frame_model = models.HybridRatedDataframe

    @staticmethod
    def init():
        migration.upgrade('head')

    def _append_time_frame(self, res_type, frame, tenant_id):
        rating_dict = frame.get('rating', {})
        rate = rating_dict.get('price')
        if not rate:
            rate = decimal.Decimal(0)
        if res_type == '_NO_DATA_':
            resource_ref = res_type
        else:
            resource_ref = frame.get('resource_id')
        if not resource_ref:
            LOG.warning('Trying to store data collected outside of '
                        'gnocchi. This driver can only be used with '
                        'the gnocchi collector. Data not stored!')
            return
        self.add_time_frame(begin=self.usage_start_dt.get(tenant_id),
                            end=self.usage_end_dt.get(tenant_id),
                            tenant_id=tenant_id,
                            res_type=res_type,
                            resource_ref=resource_ref,
                            rate=rate)

    def add_time_frame(self, **kwargs):
        """Create a new time frame.

        :param begin: Start of the dataframe.
        :param end: End of the dataframe.
        :param res_type: Type of the resource.
        :param rate: Calculated rate for this dataframe.
        :param tenant_id: tenant_id of the dataframe owner.
        :param resource_ref: Reference to the gnocchi metric (UUID).
        """
        super(GnocchiHybridStorage, self).add_time_frame(**kwargs)



# -*- coding: utf-8 -*-
# !/usr/bin/env python
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import decimal
import random

import eventlet
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_utils import uuidutils
from stevedore import driver
from tooz import coordination

from cloudkitty import collector
from cloudkitty import config  # noqa
from cloudkitty import extension_manager
from cloudkitty import messaging
from cloudkitty import storage
from cloudkitty import transformer
from cloudkitty import utils as ck_utils

eventlet.monkey_patch()

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.import_opt('backend', 'cloudkitty.tenant_fetcher', 'tenant_fetcher')

orchestrator_opts = [
    cfg.StrOpt('coordination_url',
               secret=True,
               help='Coordination driver URL',
               default='file:///var/lib/cloudkitty/locks'),
]
CONF.register_opts(orchestrator_opts, group='orchestrator')

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)

FETCHERS_NAMESPACE = 'cloudkitty.tenant.fetchers'
PROCESSORS_NAMESPACE = 'cloudkitty.rating.processors'


class RatingEndpoint(object):
    target = oslo_messaging.Target(namespace='rating',
                                   version='1.1')

    def __init__(self, orchestrator):
        self._global_reload = False
        self._pending_reload = []
        self._module_state = {}
        self._orchestrator = orchestrator

    def get_reload_list(self):
        lock = lockutils.lock('module-reload')
        with lock:
            reload_list = self._pending_reload
            self._pending_reload = []
            return reload_list

    def get_module_state(self):
        lock = lockutils.lock('module-state')
        with lock:
            module_list = self._module_state
            self._module_state = {}
            return module_list

    def quote(self, ctxt, res_data):
        LOG.debug('Received quote from RPC.')
        worker = APIWorker()
        return str(worker.quote(res_data))

    def reload_modules(self, ctxt):
        LOG.info('Received reload modules command.')
        lock = lockutils.lock('module-reload')
        with lock:
            self._global_reload = True

    def reload_module(self, ctxt, name):
        LOG.info('Received reload command for module %s.', name)
        lock = lockutils.lock('module-reload')
        with lock:
            if name not in self._pending_reload:
                self._pending_reload.append(name)

    def enable_module(self, ctxt, name):
        LOG.info('Received enable command for module %s.', name)
        lock = lockutils.lock('module-state')
        with lock:
            self._module_state[name] = True

    def disable_module(self, ctxt, name):
        LOG.info('Received disable command for module %s.', name)
        lock = lockutils.lock('module-state')
        with lock:
            self._module_state[name] = False
            if name in self._pending_reload:
                self._pending_reload.remove(name)


class BaseWorker(object):
    def __init__(self, tenant_id=None):
        self._tenant_id = tenant_id

        # Rating processors
        self._processors = []
        self._load_rating_processors()

    def _load_rating_processors(self):
        self._processors = []
        processors = extension_manager.EnabledExtensionManager(
            PROCESSORS_NAMESPACE,
            invoke_kwds={'tenant_id': self._tenant_id})

        for processor in processors:
            self._processors.append(processor)
        self._processors.sort(key=lambda x: x.obj.priority, reverse=True)


class APIWorker(BaseWorker):
    def __init__(self, tenant_id=None):
        super(APIWorker, self).__init__(tenant_id)

    def quote(self, res_data):
        for processor in self._processors:
            processor.obj.quote(res_data)

        price = decimal.Decimal(0)
        for res in res_data:
            for res_usage in res['usage'].values():
                for data in res_usage:
                    price += data.get('rating', {}).get('price',
                                                        decimal.Decimal(0))
        return price


class Worker(BaseWorker):
    def __init__(self, collector, storage, tenant_id=None):
        self._collector = collector
        if getattr(self._collector, 'collector_name') == 'ceilometer':
            LOG.warning('Ceilometer collector is deprecated and will be '
                        'removed during R cycle. This collector only works '
                        'with versions prior to Ocata')
        self._storage = storage
        self._period = METRICS_CONF['period']
        self._wait_time = METRICS_CONF['wait_periods'] * self._period

        super(Worker, self).__init__(tenant_id)

    def _collect(self, service, start_timestamp):
        next_timestamp = start_timestamp + self._period
        raw_data = self._collector.retrieve(service,
                                            start_timestamp,
                                            next_timestamp,
                                            self._tenant_id)
        if raw_data:
            return [{'period': {'begin': start_timestamp,
                                'end': next_timestamp},
                     'usage': raw_data}]

    def check_state(self):
        timestamp = self._storage.get_state(self._tenant_id)
        return ck_utils.check_time_state(timestamp,
                                         self._period,
                                         self._wait_time)

    def run(self):
        while True:
            timestamp = self.check_state()
            if not timestamp:
                break

            for service in METRICS_CONF['services']:
                try:
                    try:
                        data = self._collect(service, timestamp)
                    except collector.NoDataCollected:
                        raise
                    except Exception as e:
                        LOG.warning(
                            'Error while collecting service '
                            '%(service)s: %(error)s',
                            {'service': service, 'error': e})
                        raise collector.NoDataCollected('', service)
                except collector.NoDataCollected:
                    begin = timestamp
                    end = begin + self._period
                    for processor in self._processors:
                        processor.obj.nodata(begin, end)
                    self._storage.nodata(begin, end, self._tenant_id)
                else:
                    # Rating
                    for processor in self._processors:
                        processor.obj.process(data)
                    # Writing
                    self._storage.append(data, self._tenant_id)

            # We're getting a full period so we directly commit
            self._storage.commit(self._tenant_id)


class Orchestrator(object):
    def __init__(self):
        # Tenant fetcher
        self.fetcher = driver.DriverManager(
            FETCHERS_NAMESPACE,
            CONF.tenant_fetcher.backend,
            invoke_on_load=True).driver

        self.transformers = transformer.get_transformers()
        self.collector = collector.get_collector(self.transformers)
        self.storage = storage.get_storage(self.collector)

        # RPC
        self.server = None
        self._rating_endpoint = RatingEndpoint(self)
        self._init_messaging()

        # DLM
        self.coord = coordination.get_coordinator(
            CONF.orchestrator.coordination_url,
            uuidutils.generate_uuid().encode('ascii'))
        self.coord.start()

        self._period = METRICS_CONF['period']
        self._wait_time = METRICS_CONF['wait_periods'] * self._period

    def _lock(self, tenant_id):
        lock_name = b"cloudkitty-" + str(tenant_id).encode('ascii')
        return self.coord.get_lock(lock_name)

    def _load_tenant_list(self):
        self._tenants = self.fetcher.get_tenants()
        random.shuffle(self._tenants)

    def _init_messaging(self):
        target = oslo_messaging.Target(topic='cloudkitty',
                                       server=CONF.host,
                                       version='1.0')
        endpoints = [
            self._rating_endpoint,
        ]
        self.server = messaging.get_server(target, endpoints)
        self.server.start()

    def _check_state(self, tenant_id):
        timestamp = self.storage.get_state(tenant_id)
        return ck_utils.check_time_state(timestamp,
                                         self._period,
                                         self._wait_time)

    def process_messages(self):
        # TODO(sheeprine): Code kept to handle threading and asynchronous
        # reloading
        # pending_reload = self._rating_endpoint.get_reload_list()
        # pending_states = self._rating_endpoint.get_module_state()
        pass

    def process(self):
        while True:
            self.process_messages()
            self._load_tenant_list()
            while len(self._tenants):
                for tenant in self._tenants[:]:
                    lock = self._lock(tenant)
                    if lock.acquire(blocking=False):
                        if not self._check_state(tenant):
                            self._tenants.remove(tenant)
                        else:
                            worker = Worker(self.collector,
                                            self.storage,
                                            tenant)
                            worker.run()
                        lock.release()
                    self.coord.heartbeat()
                # NOTE(sheeprine): Slow down looping if all tenants are
                # being processed
                eventlet.sleep(1)
            # FIXME(sheeprine): We may cause a drift here
            eventlet.sleep(self._period)

    def terminate(self):
        self.coord.stop()



# -*- coding: utf-8 -*-
# Copyright 2016 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from sqlalchemy.ext import declarative

NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"}


def get_base():
    base = declarative.declarative_base()
    base.metadata.naming_convention = NAMING_CONVENTION
    return base



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import decimal

from gnocchiclient import client as gclient
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from cloudkitty import collector
from cloudkitty import utils as ck_utils


LOG = logging.getLogger(__name__)

GNOCCHI_COLLECTOR_OPTS = 'gnocchi_collector'
gnocchi_collector_opts = ks_loading.get_auth_common_conf_options()

cfg.CONF.register_opts(gnocchi_collector_opts, GNOCCHI_COLLECTOR_OPTS)
ks_loading.register_session_conf_options(
    cfg.CONF,
    GNOCCHI_COLLECTOR_OPTS)
ks_loading.register_auth_conf_options(
    cfg.CONF,
    GNOCCHI_COLLECTOR_OPTS)
CONF = cfg.CONF

METRICS_CONF = ck_utils.get_metrics_conf(CONF.collect.metrics_conf)


class GnocchiCollector(collector.BaseCollector):
    collector_name = 'gnocchi'
    dependencies = ('GnocchiTransformer',
                    'CloudKittyFormatTransformer')
    retrieve_mappings = {
        'compute': 'instance',
        'image': 'image',
        'volume': 'volume',
        'network.bw.out': 'instance_network_interface',
        'network.bw.in': 'instance_network_interface',
        'network.floating': 'network',
        'radosgw.usage': 'ceph_account',
    }
    metrics_mappings = {
        'compute': [
            ('vcpus', 'max'),
            ('memory', 'max'),
            ('cpu', 'max'),
            ('disk.root.size', 'max'),
            ('disk.ephemeral.size', 'max')],
        'image': [
            ('image.size', 'max'),
            ('image.download', 'max'),
            ('image.serve', 'max')],
        'volume': [
            ('volume.size', 'max')],
        'network.bw.out': [
            ('network.outgoing.bytes', 'max')],
        'network.bw.in': [
            ('network.incoming.bytes', 'max')],
        'network.floating': [
            ('ip.floating', 'max')],
        'radosgw.usage': [
            ('radosgw.objects.size', 'max')],
    }
    units_mappings = {
        'compute': (1, 'instance'),
        'image': ('image.size', 'MiB'),
        'volume': ('volume.size', 'GiB'),
        'network.bw.out': ('network.outgoing.bytes', 'MB'),
        'network.bw.in': ('network.incoming.bytes', 'MB'),
        'network.floating': (1, 'ip'),
        'radosgw.usage': ('radosgw.objects.size', 'GiB')
    }
    default_unit = (1, 'unknown')

    def __init__(self, transformers, **kwargs):
        super(GnocchiCollector, self).__init__(transformers, **kwargs)

        self.t_gnocchi = self.transformers['GnocchiTransformer']
        self.t_cloudkitty = self.transformers['CloudKittyFormatTransformer']

        self.auth = ks_loading.load_auth_from_conf_options(
            CONF,
            GNOCCHI_COLLECTOR_OPTS)
        self.session = ks_loading.load_session_from_conf_options(
            CONF,
            GNOCCHI_COLLECTOR_OPTS,
            auth=self.auth)
        self._conn = gclient.Client(
            '1',
            session=self.session)

    @classmethod
    def get_metadata(cls, resource_name, transformers):
        info = super(GnocchiCollector, cls).get_metadata(resource_name,
                                                         transformers)
        try:
            info["metadata"].extend(transformers['GnocchiTransformer']
                                    .get_metadata(resource_name))

            try:
                tmp = METRICS_CONF['metrics_units'][resource_name]
                info['unit'] = list(tmp.values())[0]['unit']
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                info['unit'] = cls.units_mappings[resource_name][1]

        except KeyError:
            pass
        return info

    @classmethod
    def gen_filter(cls, cop='=', lop='and', **kwargs):
        """Generate gnocchi filter from kwargs.

        :param cop: Comparison operator.
        :param lop: Logical operator in case of multiple filters.
        """
        q_filter = []
        for kwarg in sorted(kwargs):
            q_filter.append({cop: {kwarg: kwargs[kwarg]}})
        if len(kwargs) > 1:
            return cls.extend_filter(q_filter, lop=lop)
        else:
            return q_filter[0] if len(kwargs) else {}

    @classmethod
    def extend_filter(cls, *args, **kwargs):
        """Extend an existing gnocchi filter with multiple operations.

        :param lop: Logical operator in case of multiple filters.
        """
        lop = kwargs.get('lop', 'and')
        filter_list = []
        for cur_filter in args:
            if isinstance(cur_filter, dict) and cur_filter:
                filter_list.append(cur_filter)
            elif isinstance(cur_filter, list):
                filter_list.extend(cur_filter)
        if len(filter_list) > 1:
            return {lop: filter_list}
        else:
            return filter_list[0] if len(filter_list) else {}

    def _generate_time_filter(self, start, end):
        """Generate timeframe filter.

        :param start: Start of the timeframe.
        :param end: End of the timeframe if needed.
        """
        time_filter = list()
        time_filter.append(self.extend_filter(
            self.gen_filter(ended_at=None),
            self.gen_filter(cop=">=", ended_at=start),
            lop='or'))
        time_filter.append(
            self.gen_filter(cop="<=", started_at=end))
        return time_filter

    def _expand(self, metrics, resource, name, aggregate, start, end):
        try:
            values = self._conn.metric.get_measures(
                metric=metrics[name],
                start=ck_utils.ts2dt(start),
                stop=ck_utils.ts2dt(end),
                aggregation=aggregate)
            # NOTE(sheeprine): Get the list of values for the current
            # metric and get the first result value.
            # [point_date, granularity, value]
            # ["2015-11-24T00:00:00+00:00", 86400.0, 64.0]
            resource[name] = values[0][2]
        except IndexError:
            resource[name] = 0
        except KeyError:
            # Skip metrics not found
            pass

    def _expand_metrics(self, resources, mappings, start, end):
        for resource in resources:
            metrics = resource.get('metrics', {})
            try:
                for mapping in mappings:
                    self._expand(
                        metrics,
                        resource,
                        mapping.keys()[0],
                        mapping.values()[0],
                        start,
                        end,
                    )
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except AttributeError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                for name, aggregate in mappings:
                    self._expand(
                        metrics,
                        resource,
                        name,
                        aggregate,
                        start,
                        end,
                    )

    def get_resources(self, resource_name, start, end,
                      project_id, q_filter=None):
        """Get resources during the timeframe.

        :param resource_name: Resource name to filter on.
        :type resource_name: str
        :param start: Start of the timeframe.
        :param end: End of the timeframe if needed.
        :param project_id: Filter on a specific tenant/project.
        :type project_id: str
        :param q_filter: Append a custom filter.
        :type q_filter: list
        """
        # NOTE(sheeprine): We first get the list of every resource running
        # without any details or history.
        # Then we get information about the resource getting details and
        # history.

        # Translating the resource name if needed
        query_parameters = self._generate_time_filter(start, end)

        try:
            resource_type = METRICS_CONF['services_objects'].get(resource_name)
        # NOTE(mc): deprecated except part kept for backward compatibility.
        except KeyError:
            LOG.warning('Error when trying to use yaml metrology conf.')
            LOG.warning('Fallback on the deprecated oslo config method.')
            resource_type = self.retrieve_mappings.get(resource_name)

        query_parameters.append(
            self.gen_filter(cop="=", type=resource_type))
        query_parameters.append(
            self.gen_filter(project_id=project_id))
        if q_filter:
            query_parameters.append(q_filter)
        resources = self._conn.resource.search(
            resource_type=resource_type,
            query=self.extend_filter(*query_parameters))
        return resources

    def resource_info(self, resource_name, start, end, project_id,
                      q_filter=None):
        try:
            tmp = METRICS_CONF['metrics_units'][resource_name]
            qty = list(tmp.keys())[0]
            unit = list(tmp.values())[0]['unit']
        # NOTE(mc): deprecated except part kept for backward compatibility.
        except KeyError:
            LOG.warning('Error when trying to use yaml metrology conf.')
            LOG.warning('Fallback on the deprecated oslo config method.')
            qty, unit = self.units_mappings.get(
                resource_name,
                self.default_unit,
            )

        resources = self.get_resources(resource_name, start, end,
                                       project_id=project_id,
                                       q_filter=q_filter)
        formated_resources = list()
        for resource in resources:
            resource_data = self.t_gnocchi.strip_resource_data(
                resource_name, resource)

            try:
                mappings = METRICS_CONF['services_metrics'][resource_name]
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated oslo config method.')
                mappings = self.metrics_mappings[resource_name]

            self._expand_metrics([resource_data], mappings, start, end)
            resource_data.pop('metrics', None)

            # Unit conversion
            try:
                conv_data = METRICS_CONF['metrics_units'][resource_name][qty]
                if isinstance(qty, str):
                    resource_data[qty] = ck_utils.convert_unit(
                        resource_data[qty],
                        conv_data.get('factor', 1),
                        conv_data.get('offset', 0),
                    )
            # NOTE(mc): deprecated except part kept for backward compatibility.
            except KeyError:
                LOG.warning('Error when trying to use yaml metrology conf.')
                LOG.warning('Fallback on the deprecated hardcoded method.')

                if resource.get('type') == 'instance_network_interface':
                    resource_data[qty] = (
                        decimal.Decimal(resource_data[qty]) / units.M
                    )
                elif resource.get('type') == 'image':
                    resource_data[qty] = (
                        decimal.Decimal(resource_data[qty]) / units.Mi
                    )
                elif resource.get('type') == 'ceph_account':
                    resource_data[qty] = (
                        decimal.Decimal(resource_data[qty]) / units.Gi)

            data = self.t_cloudkitty.format_item(
                resource_data,
                unit,
                decimal.Decimal(
                    qty if isinstance(qty, int) else resource_data[qty]
                )
            )

            # NOTE(sheeprine): Reference to gnocchi resource used by storage
            data['resource_id'] = data['desc']['resource_id']
            formated_resources.append(data)
        return formated_resources

    def retrieve(self, resource_name, start, end, project_id, q_filter=None):
        resources = self.resource_info(resource_name, start, end,
                                       project_id=project_id,
                                       q_filter=q_filter)
        if not resources:
            raise collector.NoDataCollected(self.collector_name, resource_name)
        return self.t_cloudkitty.format_service(resource_name, resources)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from oslo_db import exception
from oslo_db.sqlalchemy import utils
from oslo_utils import uuidutils
import sqlalchemy

from cloudkitty import db
from cloudkitty.rating.hash.db import api
from cloudkitty.rating.hash.db.sqlalchemy import migration
from cloudkitty.rating.hash.db.sqlalchemy import models


def get_backend():
    return HashMap()


class HashMap(api.HashMap):

    def get_migration(self):
        return migration

    def get_service(self, name=None, uuid=None):
        session = db.get_session()
        try:
            q = session.query(models.HashMapService)
            if name:
                q = q.filter(
                    models.HashMapService.name == name)
            elif uuid:
                q = q.filter(
                    models.HashMapService.service_id == uuid)
            else:
                raise api.ClientHashMapError(
                    'You must specify either name or uuid.')
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchService(name=name, uuid=uuid)

    def get_field(self, uuid=None, service_uuid=None, name=None):
        session = db.get_session()
        try:
            q = session.query(models.HashMapField)
            if uuid:
                q = q.filter(
                    models.HashMapField.field_id == uuid)
            elif service_uuid and name:
                q = q.join(
                    models.HashMapField.service)
                q = q.filter(
                    models.HashMapService.service_id == service_uuid,
                    models.HashMapField.name == name)
            else:
                raise api.ClientHashMapError(
                    'You must specify either a uuid'
                    ' or a service_uuid and a name.')
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchField(uuid)

    def get_group(self, uuid=None, name=None):
        session = db.get_session()
        try:
            q = session.query(models.HashMapGroup)
            if uuid:
                q = q.filter(
                    models.HashMapGroup.group_id == uuid)
            if name:
                q = q.filter(
                    models.HashMapGroup.name == name)
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchGroup(name, uuid)

    def get_mapping(self, uuid):
        session = db.get_session()
        try:
            q = session.query(models.HashMapMapping)
            q = q.filter(
                models.HashMapMapping.mapping_id == uuid)
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchMapping(uuid)

    def get_threshold(self, uuid):
        session = db.get_session()
        try:
            q = session.query(models.HashMapThreshold)
            q = q.filter(
                models.HashMapThreshold.threshold_id == uuid)
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchThreshold(uuid)

    def get_group_from_mapping(self, uuid):
        session = db.get_session()
        try:
            q = session.query(models.HashMapGroup)
            q = q.join(
                models.HashMapGroup.mappings)
            q = q.filter(
                models.HashMapMapping.mapping_id == uuid)
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.MappingHasNoGroup(uuid=uuid)

    def get_group_from_threshold(self, uuid):
        session = db.get_session()
        try:
            q = session.query(models.HashMapGroup)
            q = q.join(
                models.HashMapGroup.thresholds)
            q = q.filter(
                models.HashMapThreshold.threshold_id == uuid)
            res = q.one()
            return res
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.ThresholdHasNoGroup(uuid=uuid)

    def list_services(self):
        session = db.get_session()
        q = session.query(models.HashMapService)
        res = q.values(
            models.HashMapService.service_id)
        return [uuid[0] for uuid in res]

    def list_fields(self, service_uuid):
        session = db.get_session()
        q = session.query(models.HashMapField)
        q = q.join(
            models.HashMapField.service)
        q = q.filter(
            models.HashMapService.service_id == service_uuid)
        res = q.values(models.HashMapField.field_id)
        return [uuid[0] for uuid in res]

    def list_groups(self):
        session = db.get_session()
        q = session.query(models.HashMapGroup)
        res = q.values(
            models.HashMapGroup.group_id)
        return [uuid[0] for uuid in res]

    def list_mappings(self,
                      service_uuid=None,
                      field_uuid=None,
                      group_uuid=None,
                      no_group=False,
                      **kwargs):

        session = db.get_session()
        q = session.query(models.HashMapMapping)
        if service_uuid:
            q = q.join(
                models.HashMapMapping.service)
            q = q.filter(
                models.HashMapService.service_id == service_uuid)
        elif field_uuid:
            q = q.join(
                models.HashMapMapping.field)
            q = q.filter(models.HashMapField.field_id == field_uuid)
        elif not service_uuid and not field_uuid and not group_uuid:
            raise api.ClientHashMapError(
                'You must specify either service_uuid,'
                ' field_uuid or group_uuid.')
        if 'tenant_uuid' in kwargs:
            q = q.filter(
                models.HashMapMapping.tenant_id == kwargs.get('tenant_uuid'))
        if group_uuid:
            q = q.join(
                models.HashMapMapping.group)
            q = q.filter(models.HashMapGroup.group_id == group_uuid)
        elif no_group:
            q = q.filter(models.HashMapMapping.group_id == None)  # noqa
        res = q.values(
            models.HashMapMapping.mapping_id)
        return [uuid[0] for uuid in res]

    def list_thresholds(self,
                        service_uuid=None,
                        field_uuid=None,
                        group_uuid=None,
                        no_group=False,
                        **kwargs):

        session = db.get_session()
        q = session.query(models.HashMapThreshold)
        if service_uuid:
            q = q.join(
                models.HashMapThreshold.service)
            q = q.filter(
                models.HashMapService.service_id == service_uuid)
        elif field_uuid:
            q = q.join(
                models.HashMapThreshold.field)
            q = q.filter(models.HashMapField.field_id == field_uuid)
        elif not service_uuid and not field_uuid and not group_uuid:
            raise api.ClientHashMapError(
                'You must specify either service_uuid,'
                ' field_uuid or group_uuid.')
        if 'tenant_uuid' in kwargs:
            q = q.filter(
                models.HashMapThreshold.tenant_id == kwargs.get('tenant_uuid'))
        if group_uuid:
            q = q.join(
                models.HashMapThreshold.group)
            q = q.filter(models.HashMapGroup.group_id == group_uuid)
        elif no_group:
            q = q.filter(models.HashMapThreshold.group_id == None)  # noqa
        res = q.values(
            models.HashMapThreshold.threshold_id)
        return [uuid[0] for uuid in res]

    def create_service(self, name):
        session = db.get_session()
        try:
            with session.begin():
                service_db = models.HashMapService(name=name)
                service_db.service_id = uuidutils.generate_uuid()
                session.add(service_db)
            return service_db
        except exception.DBDuplicateEntry:
            service_db = self.get_service(name=name)
            raise api.ServiceAlreadyExists(
                service_db.name,
                service_db.service_id)

    def create_field(self, service_uuid, name):
        service_db = self.get_service(uuid=service_uuid)
        session = db.get_session()
        try:
            with session.begin():
                field_db = models.HashMapField(
                    service_id=service_db.id,
                    name=name,
                    field_id=uuidutils.generate_uuid())
                session.add(field_db)
            # FIXME(sheeprine): backref are not populated as they used to be.
            #                   Querying the item again to get backref.
            field_db = self.get_field(service_uuid=service_uuid, name=name)
        except exception.DBDuplicateEntry:
            field_db = self.get_field(service_uuid=service_uuid, name=name)
            raise api.FieldAlreadyExists(field_db.name, field_db.field_id)
        else:
            return field_db

    def create_group(self, name):
        session = db.get_session()
        try:
            with session.begin():
                group_db = models.HashMapGroup(
                    name=name,
                    group_id=uuidutils.generate_uuid())
                session.add(group_db)
            return group_db
        except exception.DBDuplicateEntry:
            group_db = self.get_group(name=name)
            raise api.GroupAlreadyExists(name, group_db.group_id)

    def create_mapping(self,
                       cost,
                       map_type='rate',
                       value=None,
                       service_id=None,
                       field_id=None,
                       group_id=None,
                       tenant_id=None):
        if field_id and service_id:
            raise api.ClientHashMapError('You can only specify one parent.')
        elif not service_id and not field_id:
            raise api.ClientHashMapError('You must specify one parent.')
        elif value and service_id:
            raise api.ClientHashMapError(
                'You can\'t specify a value'
                ' and a service_id.')
        elif not value and field_id:
            raise api.ClientHashMapError(
                'You must specify a value'
                ' for a field mapping.')
        field_fk = None
        if field_id:
            field_db = self.get_field(uuid=field_id)
            field_fk = field_db.id
        service_fk = None
        if service_id:
            service_db = self.get_service(uuid=service_id)
            service_fk = service_db.id
        group_fk = None
        if group_id:
            group_db = self.get_group(uuid=group_id)
            group_fk = group_db.id
        session = db.get_session()
        try:
            with session.begin():
                field_map = models.HashMapMapping(
                    mapping_id=uuidutils.generate_uuid(),
                    value=value,
                    cost=cost,
                    field_id=field_fk,
                    service_id=service_fk,
                    map_type=map_type,
                    tenant_id=tenant_id)
                if group_fk:
                    field_map.group_id = group_fk
                session.add(field_map)
        except exception.DBDuplicateEntry:
            if field_id:
                puuid = field_id
                ptype = 'field'
            else:
                puuid = service_id
                ptype = 'service'
            raise api.MappingAlreadyExists(
                value,
                puuid,
                ptype,
                tenant_id=tenant_id)
        except exception.DBError:
            raise api.NoSuchType(map_type)
        # FIXME(sheeprine): backref are not populated as they used to be.
        #                   Querying the item again to get backref.
        field_map = self.get_mapping(field_map.mapping_id)
        return field_map

    def create_threshold(self,
                         level,
                         cost,
                         map_type='rate',
                         service_id=None,
                         field_id=None,
                         group_id=None,
                         tenant_id=None):
        if field_id and service_id:
            raise api.ClientHashMapError('You can only specify one parent.')
        elif not service_id and not field_id:
            raise api.ClientHashMapError('You must specify one parent.')
        field_fk = None
        if field_id:
            field_db = self.get_field(uuid=field_id)
            field_fk = field_db.id
        service_fk = None
        if service_id:
            service_db = self.get_service(uuid=service_id)
            service_fk = service_db.id
        group_fk = None
        if group_id:
            group_db = self.get_group(uuid=group_id)
            group_fk = group_db.id
        session = db.get_session()
        try:
            with session.begin():
                threshold_db = models.HashMapThreshold(
                    threshold_id=uuidutils.generate_uuid(),
                    level=level,
                    cost=cost,
                    field_id=field_fk,
                    service_id=service_fk,
                    map_type=map_type,
                    tenant_id=tenant_id)
                if group_fk:
                    threshold_db.group_id = group_fk
                session.add(threshold_db)
        except exception.DBDuplicateEntry:
            if field_id:
                puuid = field_id
                ptype = 'field'
            else:
                puuid = service_id
                ptype = 'service'
            raise api.ThresholdAlreadyExists(level, puuid, ptype)
        except exception.DBError:
            raise api.NoSuchType(map_type)
        # FIXME(sheeprine): backref are not populated as they used to be.
        #                   Querying the item again to get backref.
        threshold_db = self.get_threshold(threshold_db.threshold_id)
        return threshold_db

    def update_mapping(self, uuid, **kwargs):
        session = db.get_session()
        try:
            with session.begin():
                q = session.query(models.HashMapMapping)
                q = q.filter(
                    models.HashMapMapping.mapping_id == uuid)
                mapping_db = q.with_lockmode('update').one()
                if kwargs:
                    # NOTE(sheeprine): We want to check that value is not set
                    # to a None value.
                    if mapping_db.field_id and not kwargs.get('value', 'GOOD'):
                        raise api.ClientHashMapError(
                            'You must specify a value'
                            ' for a field mapping.')
                    # Resolve FK
                    if 'group_id' in kwargs:
                        group_id = kwargs.pop('group_id')
                        if group_id:
                            group_db = self.get_group(group_id)
                            mapping_db.group_id = group_db.id
                    # Service and Field shouldn't be updated
                    excluded_cols = ['mapping_id', 'service_id', 'field_id']
                    for col in excluded_cols:
                        if col in kwargs:
                            kwargs.pop(col)
                    for attribute, value in kwargs.items():
                        if hasattr(mapping_db, attribute):
                            setattr(mapping_db, attribute, value)
                        else:
                            raise api.ClientHashMapError(
                                'No such attribute: {}'.format(
                                    attribute))
                else:
                    raise api.ClientHashMapError('No attribute to update.')
                return mapping_db
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchMapping(uuid)

    def update_threshold(self, uuid, **kwargs):
        session = db.get_session()
        try:
            with session.begin():
                q = session.query(models.HashMapThreshold)
                q = q.filter(
                    models.HashMapThreshold.threshold_id == uuid)
                threshold_db = q.with_lockmode('update').one()
                if kwargs:
                    # Resolve FK
                    if 'group_id' in kwargs:
                        group_id = kwargs.pop('group_id')
                        if group_id:
                            group_db = self.get_group(group_id)
                            threshold_db.group_id = group_db.id
                    # Service and Field shouldn't be updated
                    excluded_cols = ['threshold_id', 'service_id', 'field_id']
                    for col in excluded_cols:
                        if col in kwargs:
                            kwargs.pop(col)
                    for attribute, value in kwargs.items():
                        if hasattr(threshold_db, attribute):
                            setattr(threshold_db, attribute, value)
                        else:
                            raise api.ClientHashMapError(
                                'No such attribute: {}'.format(
                                    attribute))
                else:
                    raise api.ClientHashMapError('No attribute to update.')
                return threshold_db
        except sqlalchemy.orm.exc.NoResultFound:
            raise api.NoSuchThreshold(uuid)

    def delete_service(self, name=None, uuid=None):
        session = db.get_session()
        q = utils.model_query(
            models.HashMapService,
            session)
        if name:
            q = q.filter(models.HashMapService.name == name)
        elif uuid:
            q = q.filter(models.HashMapService.service_id == uuid)
        else:
            raise api.ClientHashMapError(
                'You must specify either name or uuid.')
        r = q.delete()
        if not r:
            raise api.NoSuchService(name, uuid)

    def delete_field(self, uuid):
        session = db.get_session()
        q = utils.model_query(
            models.HashMapField,
            session)
        q = q.filter(models.HashMapField.field_id == uuid)
        r = q.delete()
        if not r:
            raise api.NoSuchField(uuid)

    def delete_group(self, uuid, recurse=True):
        session = db.get_session()
        q = utils.model_query(
            models.HashMapGroup,
            session)
        q = q.filter(models.HashMapGroup.group_id == uuid)
        with session.begin():
            try:
                r = q.with_lockmode('update').one()
            except sqlalchemy.orm.exc.NoResultFound:
                raise api.NoSuchGroup(uuid=uuid)
            if recurse:
                for mapping in r.mappings:
                    session.delete(mapping)
                for threshold in r.thresholds:
                    session.delete(threshold)
            q.delete()

    def delete_mapping(self, uuid):
        session = db.get_session()
        q = utils.model_query(
            models.HashMapMapping,
            session)
        q = q.filter(models.HashMapMapping.mapping_id == uuid)
        r = q.delete()
        if not r:
            raise api.NoSuchMapping(uuid)

    def delete_threshold(self, uuid):
        session = db.get_session()
        q = utils.model_query(
            models.HashMapThreshold,
            session)
        q = q.filter(models.HashMapThreshold.threshold_id == uuid)
        r = q.delete()
        if not r:
            raise api.NoSuchThreshold(uuid)



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import datetime
import decimal

import pecan
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from cloudkitty.api.v1.datamodels import storage as storage_models
from cloudkitty.common import policy
from cloudkitty import storage as ck_storage
from cloudkitty import utils as ck_utils


class DataFramesController(rest.RestController):
    """REST Controller to access stored data frames."""

    @wsme_pecan.wsexpose(storage_models.DataFrameCollection,
                         datetime.datetime,
                         datetime.datetime,
                         wtypes.text,
                         wtypes.text)
    def get_all(self, begin=None, end=None, tenant_id=None,
                resource_type=None):
        """Return a list of rated resources for a time period and a tenant.

        :param begin: Start of the period
        :param end: End of the period
        :param tenant_id: UUID of the tenant to filter on.
        :param resource_type: Type of the resource to filter on.
        :return: Collection of DataFrame objects.
        """

        policy.authorize(pecan.request.context, 'storage:list_data_frames', {})

        if not begin:
            begin = ck_utils.get_month_start()
        if not end:
            end = ck_utils.get_next_month()

        begin_ts = ck_utils.dt2ts(begin)
        end_ts = ck_utils.dt2ts(end)
        backend = pecan.request.storage_backend
        dataframes = []
        try:
            frames = backend.get_time_frame(begin_ts,
                                            end_ts,
                                            tenant_id=tenant_id,
                                            res_type=resource_type)
            for frame in frames:
                for service, data_list in frame['usage'].items():
                    frame_tenant = None
                    resources = []
                    for data in data_list:
                        desc = data['desc'] if data['desc'] else {}
                        price = decimal.Decimal(str(data['rating']['price']))
                        resource = storage_models.RatedResource(
                            service=service,
                            desc=desc,
                            volume=data['vol']['qty'],
                            rating=price)
                        frame_tenant = data['tenant_id']
                        resources.append(resource)
                    dataframe = storage_models.DataFrame(
                        begin=ck_utils.iso2dt(frame['period']['begin']),
                        end=ck_utils.iso2dt(frame['period']['end']),
                        tenant_id=frame_tenant,
                        resources=resources)
                    dataframes.append(dataframe)
        except ck_storage.NoTimeFrame:
            pass
        return storage_models.DataFrameCollection(dataframes=dataframes)


class StorageController(rest.RestController):
    """REST Controller to access stored data."""

    dataframes = DataFramesController()



# -*- coding: utf-8 -*-
# Copyright 2015 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
import csv
import json

from oslo_config import cfg

from cloudkitty import collector

fake_collector_opts = [
    cfg.StrOpt('file',
               default='/var/lib/cloudkitty/input.csv',
               help='Collector input file.')]

cfg.CONF.register_opts(fake_collector_opts, 'fake_collector')


class CSVCollector(collector.BaseCollector):
    collector_name = 'csvcollector'
    dependencies = ('CloudKittyFormatTransformer', )

    def __init__(self, transformers, **kwargs):
        super(CSVCollector, self).__init__(transformers, **kwargs)

        self.t_cloudkitty = self.transformers['CloudKittyFormatTransformer']
        self._file = None
        self._csv = None

    def _open_csv(self):
        filename = cfg.CONF.fake_collector.file
        csvfile = open(filename, 'rb')
        reader = csv.DictReader(csvfile)
        self._file = csvfile
        self._csv = reader

    @classmethod
    def get_metadata(cls, resource_name, transformers):
        res = super(CSVCollector, cls).get_metadata(resource_name,
                                                    transformers)
        try:
            filename = cfg.CONF.fake_collector.file
            csvfile = open(filename, 'rb')
            reader = csv.DictReader(csvfile)
            entry = None
            for row in reader:
                if row['type'] == resource_name:
                    entry = row
                    break
            res['metadata'] = json.loads(entry['desc']).keys() if entry else {}
        except IOError:
            pass
        return res

    def filter_rows(self,
                    start,
                    end=None,
                    project_id=None,
                    res_type=None):
        rows = []
        for row in self._csv:
            if int(row['begin']) == start:
                if res_type:
                    if row['type'] == res_type:
                        rows.append(row)
                else:
                    rows.append(row)
        return rows

    def _get_data(self,
                  res_type,
                  start,
                  end=None,
                  project_id=None,
                  q_filter=None):
        self._open_csv()
        rows = self.filter_rows(start, end, project_id, res_type=res_type)
        data = []
        for row in rows:
            data.append({
                'desc': json.loads(row['desc']),
                'vol': json.loads(row['vol'])})
        if not data:
            raise collector.NoDataCollected(self.collector_name, res_type)
        return self.t_cloudkitty.format_service(res_type, data)

    def get_compute(self,
                    start,
                    end=None,
                    project_id=None,
                    q_filter=None):
        return self._get_data('compute',
                              start,
                              end,
                              project_id,
                              q_filter)

    def get_image(self,
                  start,
                  end=None,
                  project_id=None,
                  q_filter=None):
        return self._get_data('image',
                              start,
                              end,
                              project_id,
                              q_filter)

    def get_volume(self,
                   start,
                   end=None,
                   project_id=None,
                   q_filter=None):
        return self._get_data('volume',
                              start,
                              end,
                              project_id,
                              q_filter)

    def get_network_bw_in(self,
                          start,
                          end=None,
                          project_id=None,
                          q_filter=None):
        return self._get_data('network.bw.in',
                              start,
                              end,
                              project_id,
                              q_filter)

    def get_network_bw_out(self,
                           start,
                           end=None,
                           project_id=None,
                           q_filter=None):
        return self._get_data('network.bw.out',
                              start,
                              end,
                              project_id,
                              q_filter)

    def get_network_floating(self,
                             start,
                             end=None,
                             project_id=None,
                             q_filter=None):
        return self._get_data('network.floating',
                              start,
                              end,
                              project_id,
                              q_filter)



# -*- coding: utf-8 -*-
# Copyright 2016 99Cloud zhangguoqing <zhang.guoqing@99cloud.net>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
import oslo_messaging
from oslo_messaging.rpc import dispatcher

DEFAULT_URL = "__default__"
RPC_TARGET = None
TRANSPORTS = {}


def setup():
    oslo_messaging.set_transport_defaults('cloudkitty')


def get_transport(url=None, optional=False, cache=True):
    """Initialise the oslo_messaging layer."""
    global TRANSPORTS, DEFAULT_URL
    cache_key = url or DEFAULT_URL
    transport = TRANSPORTS.get(cache_key)
    if not transport or not cache:
        try:
            transport = oslo_messaging.get_rpc_transport(cfg.CONF, url)
        except (oslo_messaging.InvalidTransportURL,
                oslo_messaging.DriverLoadFailure):
            if not optional or url:
                # NOTE(sileht): oslo_messaging is configured but unloadable
                # so reraise the exception
                raise
            return None
        else:
            if cache:
                TRANSPORTS[cache_key] = transport
    return transport


def get_target():
    global RPC_TARGET
    if RPC_TARGET is None:
        RPC_TARGET = oslo_messaging.Target(topic='cloudkitty', version='1.0')
    return RPC_TARGET


def get_client(version_cap=None):
    transport = get_transport()
    target = get_target()
    return oslo_messaging.RPCClient(transport, target, version_cap=version_cap)


def get_server(target=None, endpoints=None):
    access_policy = dispatcher.DefaultRPCAccessPolicy
    transport = get_transport()
    if not target:
        target = get_target()
    return oslo_messaging.get_rpc_server(transport, target,
                                         endpoints, executor='eventlet',
                                         access_policy=access_policy)


def cleanup():
    """Cleanup the oslo_messaging layer."""
    global TRANSPORTS, NOTIFIERS
    NOTIFIERS = {}
    for url in TRANSPORTS:
        TRANSPORTS[url].cleanup()
        del TRANSPORTS[url]



# -*- coding: utf-8 -*-
# Copyright 2014 Objectif Libre
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# @author: Stéphane Albert
#
from cloudkitty import service


def main():
    service.prepare_service()

    # NOTE(mc): This import is done here to ensure that the prepare_service()
    # function is called before any cfg option. By importing the orchestrator
    # file, the utils one is imported too, and then some cfg options are read
    # before the prepare_service(), making cfg.CONF returning default values
    # systematically.
    from cloudkitty import orchestrator
    processor = orchestrator.Orchestrator()
    try:
        processor.process()
    except KeyboardInterrupt:
        processor.terminate()


if __name__ == '__main__':
    main()
